import guidance
from ._block import block

@guidance
def role_opener(lm, role_name, **kwargs):
    if not hasattr(lm, "get_role_start"):
        raise Exception(f"You need to use a chat model in order the use role blocks like `with {role_name}():`! Perhaps you meant to use the {type(lm).__name__}Chat class?")
    lm += f"<||_html:<div style='display: flex; border-bottom: 1px solid rgba(127, 127, 127, 0.2); align-items: center;'><div style='flex: 0 0 80px; opacity: 0.5;'>{role_name.lower()}</div><div style='flex-grow: 1; padding: 5px; padding-top: 10px; padding-bottom: 10px; margin-top: 0px; white-space: pre-wrap; margin-bottom: 0px;'>_||>"
    lm += "<||_#NODISP_||>" + lm.get_role_start(role_name, **kwargs) + "<||_/NODISP_||>"
    return lm

@guidance
def role_closer(lm, role_name, **kwargs):
    lm += "<||_html:</div></div>_||>" + "<||_#NODISP_||>" + lm.get_role_end(role_name) + "<||_/NODISP_||>"
    return lm

def role(role_name, text=None, **kwargs):
    if text is None:
        return block(opener=role_opener(role_name, **kwargs), closer=role_closer(role_name, **kwargs))
    else:
        assert False
        #return self.append(open_text + text + close_text)

def system(text=None, **kwargs):
    return role("system", text, **kwargs)

def user(text=None, **kwargs):
    return role("user", text, **kwargs)

def assistant(text=None, **kwargs):
    return role("assistant", text, **kwargs)

def function(text=None, **kwargs):
    return role("function", text, **kwargs)

def instruction(text=None, **kwargs):
    return role("instruction", text, **kwargs)


import pytest
import guidance
from guidance import gen, capture, select, user, system, assistant
from ..utils import get_model

def test_openai_basic():
    try:
        lm = guidance.models.OpenAI("text-curie-001")
    except:
        pytest.skip("Skipping OpenAI test because we can't load the model!")
    # lm = guidance.models.Transformers("gpt2")
    lm += "Count to 20: 1,2,3,4,"
    nl = "\n"
    lm += f"""\
    5,6,7"""
    lm += f"""{gen(max_tokens=1, suffix=nl)}aaaaaa"""
    assert str(lm)[-5:] == "aaaaa"

def test_openai_select():
    try:
        lm = guidance.models.OpenAI("text-curie-001")
    except:
        pytest.skip("Skipping OpenAI test because we can't load the model!")
    lm += "Pick a number: "
    lm += select(["1", "11", "111", "1111", "11111", "111111", "1111111"], name='the number')
    assert str(lm)[-1] in "123"

def test_openai_chat():
    try:
        lm = guidance.models.OpenAI("gpt-3.5-turbo")
    except:
        pytest.skip("Skipping OpenAI test because we can't load the model!")
    with system():
        lm += "You are a math wiz."

    with user():
        lm += "What is 1 + 1?"

    with assistant():
        lm += gen(max_tokens=10, name="text")
        lm += "Pick a number: "

    assert len(lm["text"]) > 0


import os
import guidance
import pytest

opanai_model_cache = {}

def get_model(model_name, caching=False, **kwargs):
    """ Get an LLM by name.
    """
    if model_name.startswith("openai:"):
        return get_openai_model(model_name[7:], caching, **kwargs)
    elif model_name.startswith("transformers:"):
        return get_transformers_model(model_name[13:], caching, **kwargs)
    elif model_name.startswith("llama_cpp:"):
        return get_llama_cpp_model(model_name[10:], caching, **kwargs)

def get_openai_model(model_name, caching=False, **kwargs):
    """ Get an OpenAI LLM with model reuse and smart test skipping.
    """

    # we cache the models so lots of tests using the same model don't have to
    # load it over and over again
    key = model_name+"_"+str(caching)
    if key not in opanai_model_cache:
        opanai_model_cache[key] = guidance.llms.OpenAI(model_name, caching=caching, **kwargs)
    llm = opanai_model_cache[key]

    if llm.api_key is None:
        pytest.skip("OpenAI token not found")

    return llm

transformers_model_cache = {}

def get_transformers_model(model_name, caching=False):
    """ Get an OpenAI LLM with model reuse.
    """

    # we cache the models so lots of tests using the same model don't have to
    # load it over and over again
    key = model_name+"_"+str(caching)
    if key not in transformers_model_cache:
        transformers_model_cache[key] = guidance.models.Transformers(model_name, caching=caching)

    return transformers_model_cache[key]

llama_cpp_model_cache = {}

def get_llama_cpp_model(model_name, caching=False):
    """ Get a llama.cpp LLM with model reuse.
    """

    if model_name is None or isinstance(model_name, str) and len(model_name.strip()) == 0:
        model_name = os.environ.get("LLAMA_CPP_MODEL", "")
        if len(model_name.strip()) == 0:
            pytest.skip("No llama_cpp model found.")

    # we cache the models so lots of tests using the same model don't have to
    # load it over and over again
    key = model_name+"_"+str(caching)
    if key not in llama_cpp_model_cache:
        llama_cpp_model_cache[key] = guidance.models.LlamaCpp(model_name, caching=caching)

    return llama_cpp_model_cache[key]



import base64
import uuid
import json
import inspect
import types
import re

tag_start = "{{G|"
tag_end = "|G}}"
_call_pool = {}
_tag_pattern = re.compile(re.escape(tag_start) + r"([^\|]+)" + re.escape(tag_end))

class StatefulException(Exception):
    '''This is raised when we try and use the state of a grammar object like it was a live model.
    
    Note that eventually we do want to support stateful parser/grammar constructs directly, but
    for now we use a traditional parser and grammar separation (hence the need for this exception).'''
    pass

_excluded_names = frozenset(["_find_name", "__add__", "__radd__", "<listcomp>", "<module>", "select", "char_set"])
def _find_name():
    stack = inspect.stack()
    for frame in stack:
        name = frame.function
        if frame.function not in _excluded_names:
            return name
    return None

class Function():
    def __init__(self, name, value=None) -> None:
        self.name = name
        self.value = value

    def __str__(self):
        '''Creates a string tag that can be used to retrieve this object.'''
    
        # save the call in our call pool, ready to be run when it is attached to an LM object
        str_id = str(id(self))
        if str_id not in _call_pool:
            _call_pool[str_id] = self

        # return a string representation of this call so it can be combined with other strings/calls
        return tag_start + str_id + tag_end

class StatefulFunction(Function):
    __slots__ = ("f", "args", "kwargs")

    def __init__(self, f, args, kwargs):
        self.f = f
        self.args = args
        self.kwargs = kwargs
    
    def __call__(self, model):
        return self.f(model, *self.args, **self.kwargs)
    
    def __add__(self, other):
        
        # if we are joining with a string we use the string representation for ourselves
        if isinstance(other, str):
            return str(self) + other
        
        def __add__(model):
            model = self(model)
            if model is None:
                raise Exception(f"The guidance function `{self.f.__name__}` did not return a model object! You need to return an updated model object at the end of your guidance function.")
            if isinstance(other, StatelessFunction):
                return model + other
            else:
                return other(model)
        return StatefulFunction(__add__, [], {})
    
    def __radd__(self, other):
        
        # if we are joining with a string we use the string representation for ourselves
        if isinstance(other, str):
            return other + str(self)
        
        def __radd__(model):
            if isinstance(other, StatelessFunction):
                model += other
            else:
                model = other(model)
            return self(model)
        return StatefulFunction(__radd__, [], {})

class StatelessFunction(Function):
    num_used_names = 0

    def __add__(self, value):

        # see if we have a string with calls or a simple string 
        if isinstance(value, str) or isinstance(value, bytes):
            if isinstance(value, str) and re.search(_tag_pattern, value):
                return str(self) + value
            else:
                value = string(value) 
        
        # see if we can keep building a stateless grammar
        if isinstance(value, StatelessFunction):
            return Join([self, value])
        
        # otherwise we let the stateful object handle things
        else:
            return value.__radd__(self)
    
    def __radd__(self, value):

        # see if we have a string with calls or a simple string 
        if isinstance(value, str) or isinstance(value, bytes):
            if isinstance(value, str) and re.search(_tag_pattern, value):
                return value + str(self)
            else:
                value = string(value) 
        
        # see if we can keep building a stateless grammar
        if isinstance(value, StatelessFunction):
            return Join([value, self])
        
        # otherwise we let the stateful object handle things
        else:
            return value.__add__(self)
    
    def __getitem__(self, value):
        raise StatefulException("StatelessFunctions can't access state!")
    
    @staticmethod
    def _new_name():
        num_used = StatelessFunction.num_used_names

        a_ord = ord('a')

        # name the name in base 26 letter notation
        name = chr(a_ord + num_used % 26)
        if num_used >= 26:
            name = chr(a_ord + (num_used % 676) // 26) + name
            if num_used >= 676:
                name = chr(a_ord + (num_used % 17576) // 676) + name
                if num_used >= 17576:
                    name = chr(a_ord + (num_used % 456976) // 17576) + name

        StatelessFunction.num_used_names += 1
        
        return name
    
    def gbnf_string(self):
        used_names = set()
        names = {}
        lines = []
        root_name = self._rec_gbnf_string(lines, used_names, names)
        lines.append("root ::= " + root_name)
        return "\n".join(lines)

class Terminal(StatelessFunction):
    def match_byte(self, byte):
        pass # abstract

    @property
    def max_tokens(self):
        return 1000000000000

class Byte(Terminal):
    __slots__ = ("byte", "hidden", "commit_point", "capture_name", "temperature")

    def __init__(self, byte):
        assert isinstance(byte, bytes)
        assert len(byte) == 1
        self.byte = byte
        self.hidden = False
        self.commit_point = False
        self.capture_name = None
        self.temperature = -1

    @property
    def name(self):
        return str(self.byte)
    
    def __hash__(self):
        return self.byte[0]
    
    def __eq__(self, other):
        return isinstance(other, Byte) and self.byte[0] == other.byte[0]
    
    def __repr__(self) -> str:
        return str(self.byte)
    
    def __len__(self):
        return 1
    
    def match_byte(self, byte):
        return byte == self.byte
    
    @property
    def nullable(self):
        return False

class ByteRange(Terminal):
    __slots__ = ("byte_range", "hidden", "commit_point", "capture_name", "temperature")

    def __init__(self, byte_range):
        assert isinstance(byte_range, bytes)
        assert len(byte_range) == 2
        self.byte_range = byte_range
        self.hidden = False
        self.commit_point = False
        self.capture_name = None
        self.temperature = -1 # -1 means not set

    def match_byte(self, byte):
        return self.byte_range[0] <= byte[0] <= self.byte_range[1]

    @property
    def name(self):
        return str(self.byte_range)
    @name.setter
    def name(self, value):
        pass # we ignore name changes
    
    @property
    def nullable(self):
        return False
    
    def __hash__(self):
        return self.byte_range[0] + 256 * self.byte_range[1]
    
    def __eq__(self, other):
        return isinstance(other, ByteRange) and self.byte_range[0] == other.byte_range[0] and self.byte_range[1] == other.byte_range[1]
    
    def __repr__(self) -> str:
        return str(self.byte_range)
    
    def __len__(self):
        return 1

class Null():
    __slots__ = ("name", "hidden", "commit_point", "capture_name")

    nullable = True
    def __init__(self):
        self.name = None
        self.hidden = False
        self.commit_point = False
        self.capture_name = None

    def __add__(self, other):
        if isinstance(other, str):
            return string(other)
        else:
            return other
        
    def __radd__(self, other):
        if isinstance(other, str):
            return string(other)
        else:
            return other
        
class ModelVariable(StatelessFunction):
    '''This represents a variable that will be read from the model object when this grammar is executed.
    
    Note that the name is the name of the attribute on the model object this node
    will get replaced with.
    '''
    __slots__ = ("name", "hidden", "commit_point", "capture_name")

    def __init__(self, name):
        self.name = name
        self.hidden = False
        self.commit_point = False
        self.capture_name = None
        self.nullable = False

def replace_grammar_node(grammar, target, replacement):
    # Use a stack to keep track of the nodes to be visited
    stack = [grammar]
    visited_set = set()  # use set for O(1) lookups

    while stack:
        current = stack.pop()
        
        # Check if we have already visited this node
        if current in visited_set:
            continue
        visited_set.add(current)

        # We are done with this node if it's a terminal
        if isinstance(current, (Terminal, ModelVariable)):
            continue
        
        # Iterate through the node's values and replace target with replacement
        for i, value in enumerate(current.values):
            if value == target:
                current.values[i] = replacement
            else:
                stack.append(value)

# def replace_grammar_node(grammar, target, replacement, visited_set={}):
    
#     # see if we have already visited this node
#     if grammar in visited_set:
#         return
#     else:
#         visited_set[grammar] = True
   
#     # we are done if this is a terminal
#     if isinstance(grammar, (Terminal, ModelVariable)):
#         return
    
#     # replace all matching sub-nodes
#     for i,value in enumerate(grammar.values):
#         if value == target:
#             grammar.values[i] = replacement
#         else:
#             replace_grammar_node(value, target, replacement, visited_set)

def replace_model_variables(grammar, model):
    '''Replace all the ModelVariable nodes with their values in an iterative manner.'''
    visited_set = set()
    stack = [(grammar, None, None)]  # Stack stores tuples of (node, parent_node, child_index)
    replacements = []

    while stack:
        current, parent, child_index = stack.pop()

        # This node is being visited for the first time
        if current not in visited_set:
            visited_set.add(current)

            # If it's a terminal node, skip it
            if isinstance(current, Terminal):
                continue

            # Process non-terminal nodes in reverse order to maintain the depth-first order
            for i in reversed(range(len(current.values))):
                value = current.values[i]
                if isinstance(value, ModelVariable):
                    # Replace the ModelVariable with its value from 'model'
                    replacement_value = _wrap_as_grammar(getattr(model, value.name))
                    if value.commit_point:
                        replacement_value = commit_point(replacement_value, hidden=value.hidden)
                    replacements.append((current, i, value))  # Record the replacement
                    current.values[i] = replacement_value  # Perform the replacement
                else:
                    # If not ModelVariable, push onto the stack to process later
                    stack.append((value, current, i))

    return replacements

# def replace_model_variables(grammar, model, visited_set={}):
#     '''Replace all the ModelVariable nodes with their values.'''
    
#     # see if we have already visited this node
#     if grammar in visited_set:
#         return []
#     else:
#         visited_set[grammar] = True
   
#     # we are done if this is a terminal
#     if isinstance(grammar, Terminal):
#         return []
    
#     # replace all matching sub-nodes
#     replacements = []
#     for i,value in enumerate(grammar.values):
#         if isinstance(value, ModelVariable):
#             g = _wrap_as_grammar(getattr(model, value.name))
#             if value.commit_point:
#                 g = commit_point(g, hidden=value.hidden)
#             replacements.append((grammar, i, value))
#             grammar.values[i] = g
#         else:
#             replacements.extend(replace_model_variables(value, model, visited_set))
#     return replacements

def unreplace_model_variables(replacements):
    '''This restores a grammar back to its original state, ready for another execution.'''
    for grammar,i,orig_value in replacements:
        grammar.values[i] = orig_value

def _wrap_as_grammar(value):
    '''This takes whatever value was given and tries to turn in into a guidance grammar.'''

    # if it is already a valid grammar we have no need to wrap it
    if isinstance(value, StatelessFunction):
        return value
    
    # if it is already a valid grammar we have no need to wrap it
    if value is None:
        return Null() 
    
    # we have a constant value
    if isinstance(value, (str, bytes)):
        return string(value)
    
    raise Exception("Can't wrap as a grammar!")

def commit_point(value, hidden=False):
    '''Force the grammar to commit to a parse that includes this node once it can.
    
    Not that commit point nodes can be optionally hidden (in fact they are the only
    nodes that can be hidden since they are by definition not impacted by multiple possible
    inconsistent parses.)'''
    # TODO: assert that value is not empty since we don't yet support that
    if isinstance(value, str):
        value = string(value)
    if isinstance(value, Terminal):
        value = Join([value]) # commit points should be full nodes (otherwise we can't hide them) TODO: decide if we want to do this even for non-hidden commit points
    value.commit_point = True
    if hidden:
        _rec_hide(value)
    return value

def _rec_hide(grammar):
    if not grammar.hidden:
        grammar.hidden = True
        if hasattr(grammar, "values"):
            for g in grammar.values:
                _rec_hide(g)

class Placeholder(StatelessFunction):
    __slots__ = tuple("nullable")
    def __init__(self):
        self.nullable = False


class Join(StatelessFunction):
    __slots__ = ("nullable", "values", "name", "hidden", "commit_point", "capture_name", "max_tokens")

    def __init__(self, values, name=None, max_tokens=100000000) -> None:
        self.nullable = all(v.nullable for v in values)
        self.values = [v for v in values if not isinstance(v, Null)]
        self.name = name if name is not None else StatelessFunction._new_name()
        self.hidden = False
        self.commit_point = False
        self.capture_name = None
        self.max_tokens = max_tokens

    def __repr__(self, indent="", done=None):
        if done is None:
            done = set()
        s = self.name.ljust(20) + " <- " + " ".join([v.name for v in self.values])
        s += "        " + ("hidden " if self.hidden else "") + ("commit_point " if self.commit_point else "") + (f"max_tokens={self.max_tokens}" if self.max_tokens < 100000 else "") +"\n"
        done.add(self)
        for v in self.values:
            if v not in done and (isinstance(v, Join) or isinstance(v, Select)):
                s += v.__repr__(indent, done)
        return s

class Select(StatelessFunction):
    __slots__ = ("nullable", "_values", "name", "hidden", "commit_point", "capture_name", "max_tokens", "recursive")

    def __init__(self, values, capture_name=None, name=None, max_tokens=10000000, recursive=False) -> None:
        self.values = values
        self.name = name if name is not None else StatelessFunction._new_name()
        self.hidden = False
        self.commit_point = False
        self.capture_name = capture_name
        self.max_tokens = max_tokens
        self.recursive = recursive

    @property
    def values(self):
        return self._values
    @values.setter
    def values(self, vals):
        self._values = [string(v) if isinstance(v, (str, bytes)) else v for v in vals]
        self.nullable = any(v.nullable for v in self._values)
        self._values = [v for v in self._values if not isinstance(v, Null)]

    def __repr__(self, indent="", done=None):
        if done is None:
            done = set()
        s = self.name.ljust(20) + " <- " + " | ".join([v.name for v in self.values])
        s += "        " + ("hidden " if self.hidden else "") + ("commit_point " if self.commit_point else "") + (f"max_tokens={self.max_tokens}" if self.max_tokens < 100000 else "") +"\n"
        done.add(self)
        for v in self.values:
            if v not in done and (isinstance(v, Join) or isinstance(v, Select)):
                s += v.__repr__(indent, done)
        return s

def string(value):
    if isinstance(value, str):
        b = bytes(value, encoding="utf8")
    elif isinstance(value, bytes):
        b = value
    else:
        raise Exception("Must pass bytes or str to the string() function!")
    if len(value) == 0:
        return Null()
    elif len(b) == 1:
        return Byte(b)
    else:
        return Join([Byte(b[i:i+1]) for i in range(len(b))], name=str(b))
    
def select(options, name=None, recurse=False, skip_checks=False):
    # TODO: allow for returning the probabilites of the selected item
    # TODO: also the full probabilites distribution over all items. We can implement this using the prob of the selected item by repeating the call, removing the selected item each time
    if not skip_checks:
        for i, value in enumerate(options):
            assert not isinstance(value, StatefulFunction), "You cannot select between stateful functions in the current guidance implementation!"
            assert not isinstance(value, types.FunctionType), "Did you pass a function without calling it to select? You need to pass the results of a called guidance function to select."
            if isinstance(value, int) or isinstance(value, float):
                options[i] = str(value)
    # if name is None:
    #     name = _find_name() + "_" + StatelessFunction._new_name()
    if recurse:
        node = Select([], capture_name=name, recursive=True)
        node.values = [node + v for v in options if v != ""] + options
        return node
    else:
        if len(options) == 1 and name is None:
            return options[0]
        else:
            return Select(options, capture_name=name, recursive=False)
        
def byte_range(low, high):
    return ByteRange(low + high)

# def ignore_placeholders(value):
#     if not isinstance(value, Join): # don't double wrap
#         value = Join([value]) # this ensures we capture what we want, and not something surprisingly self_recursive
#     value.ignore_placeholders = True
#     return value

def capture(value, name):
    if not (isinstance(value, Join) and len(value.values) == 1): # don't double wrap
        value = Join([value]) # this ensures we capture what we want, and not something surprisingly self_recursive
    value.capture_name = name
    return value

def token_limit(value, max_tokens):
    _rec_token_limit(value, max_tokens)
    return value

def _rec_token_limit(grammar, max_tokens):
    if grammar.max_tokens > max_tokens and not isinstance(grammar, Terminal):
        if getattr(grammar, "recursive", False): # only restrict recursive selects, otherwise we would block all way to complete the grammar
            grammar.max_tokens = max_tokens
        if hasattr(grammar, "values"):
            for g in grammar.values:
                _rec_token_limit(g, max_tokens)

def with_temperature(value, temperature):
    '''This sets the sampling temperature to be used for the given portion of the grammar.
    
    Note that if the grammar passed to us already has some portions with a temperature
    setting in place, those setting will not be overridden.
    '''
    _re_with_temperature(value, temperature, {})
    return value

def _re_with_temperature(grammar, temperature, visited_set):
    
    # don't go down the same path twice
    if grammar in visited_set:
        return
    visited_set[grammar] = True

    # if getattr(grammar, "temperature", 100000000) > temperature:
    if isinstance(grammar, Terminal) and grammar.temperature < 0: # only need to set temp for terminals
        grammar.temperature = temperature
    elif getattr(grammar, "temperature", 100000000) > temperature and hasattr(grammar, "values"):
        for g in grammar.values:
            _re_with_temperature(g, temperature, visited_set)

def model_variable(name):
    return ModelVariable(name)

# def char_range(low, high):
#     low_bytes = bytes(low, encoding="utf8")
#     high_bytes = bytes(high, encoding="utf8")
#     if len(low_bytes) > 1 or len(high_bytes) > 1:
#         raise Exception("We don't yet support multi-byte character ranges!")
#     return ByteRange(low_bytes + high_bytes)


from .._model import Chat, Instruct
# from ._local import Local
from .._remote import Remote


try:
    # TODO: can we eliminate the torch requirement for llama.cpp by using numpy in the caller instead?
    import torch
    is_torch = True
except ImportError:
    is_torch = False

try:
    # TODO: can we eliminate the torch requirement for llama.cpp by using numpy in the caller instead?
    import vertexai
    is_vertexai = True
except ImportError:
    is_vertexai = False

class VertexAI(Remote):
    def __init__(self, model, tokenizer=None, echo=True, caching=True, temperature=0.0, max_streaming_tokens=500, **kwargs):
        if not is_vertexai:
            raise Exception("Please install the vertexai package using `pip install google-cloud-aiplatform` in order to use guidance.models.VertexAI!")
        
        # if we are called directly (as opposed to through super()) then we convert ourselves to a more specific subclass if possible
        if self.__class__ is VertexAI:
            found_subclass = None
            from .. import vertexai

            if isinstance(model, str):
                model_name = model
            else:
                model_name = self.model_obj._model_id

            # PaLM2Instruct
            if model_name == "text-bison@001":
                found_subclass = vertexai.PaLM2Instruct

            # PaLM2Chat
            elif model_name == "chat-bison@001":
                found_subclass = vertexai.PaLM2Chat
            
            # convert to any found subclass
            if found_subclass is not None:
                self.__class__ = found_subclass
                found_subclass.__init__(self, model, tokenizer=tokenizer, echo=echo, caching=caching, temperature=temperature, max_streaming_tokens=max_streaming_tokens, **kwargs)
                return # we return since we just ran init above and don't need to run again
        
            # make sure we have a valid model object
            if isinstance(model, str):
                raise Exception("The model ID you passed, `{model}`, does not match any known subclasses!")

        super().__init__(
            model, tokenizer=tokenizer, echo=echo,
            caching=caching, temperature=temperature,
            max_streaming_tokens=max_streaming_tokens, **kwargs
        )
        # self.caching = caching
        # self.temperature = temperature
        # self.max_streaming_tokens = max_streaming_tokens

        # # Vertex AI models don't always have public tokenizations, so when not provided we pretend they tokenize like gpt2...
        # if tokenizer is None:
        #     tokenizer = tiktoken.get_encoding("gpt2") # TODO: auto detect some common models in the model garden

        # # tiktoken tokenizer was given
        # if hasattr(tokenizer, "decode_single_token_bytes"):
        #     byte_tokens = [tokenizer.decode_single_token_bytes(i) for i in range(tokenizer.n_vocab)] + [b"<|endofprompt|>"]
        #     bos_token_id = None
        #     eos_token_id = len(byte_tokens)-2
        
        # # a transformer tokenizer was given that has a byte_decoder
        # elif hasattr(tokenizer, "byte_decoder"):
        #     byte_tokens = []
        #     for i in range(tokenizer.vocab_size):
        #         byte_coded = bytes([tokenizer.byte_decoder[c] for c in tokenizer.convert_ids_to_tokens(i)])
        #         byte_tokens.append(byte_coded)
        #     bos_token_id = tokenizer.bos_token_id
        #     eos_token_id = tokenizer.eos_token_id
        
        # # a transformer tokenizer was given with byte_decoder
        # elif hasattr(tokenizer, "convert_ids_to_tokens"):
        #     byte_tokens = [bytes(tokenizer.convert_tokens_to_string(['a', tokenizer.convert_ids_to_tokens(i)])[1:], encoding="utf8") for i in range(tokenizer.vocab_size)]
        #     bos_token_id = tokenizer.bos_token_id
        #     eos_token_id = tokenizer.eos_token_id

        # else:
        #     raise Exception("The tokenizer given was not of a recognized type!")

        # # build the 
        # super().__init__(
        #     byte_tokens,
        #     bos_token_id,
        #     eos_token_id,
        #     echo=echo
        # )

        # self._shared_state = {
        #     "data_queue": queue.Queue(), # this is where the streaming thread puts results
        #     "data": b'', # these are the bytes we are ready to use in the main thread
        #     "not_running_stream": threading.Event(), # this is phrased negatively so we can wait for the stop event
        #     "last_call": 0,
        #     "num_calls_made": 0
        # }
        # self._shared_state["not_running_stream"].set()

        # self.max_calls = 10
        # self.timeout = 5.1

    # def _running_stream(self):
    #     return not self._shared_state["not_running_stream"].is_set() # wrap double negation (which)

    # def _start_generator_stream(self, generator):
    #     dqueue = self._shared_state["data_queue"]
    #     first_iteration = True
    #     for chunk in generator:
    #         bchunk = chunk.text.encode("utf8")
    #         if len(bchunk) > 0:
    #             dqueue.put(bchunk)
    #         if self._shared_state["not_running_stream"].is_set() or not first_iteration and time.time() - self._shared_state["last_call"] > self.timeout:
    #             break
    #         first_iteration = False

    #     if self._running_stream():
    #         dqueue.put(self.tokens[self.eos_token_id])
    #     self._shared_state["not_running_stream"].set()
    #     dqueue.put(b'') # so we never get stuck waiting for a running stream to return something
        
    # def _start_new_stream(self, prompt):

    #     if self._shared_state["num_calls_made"] > self.max_calls:
    #         raise Exception(f"We have exceeded the maximum number of calls! {self.max_calls}")


    #     # stop any running stream
    #     if self._running_stream():
    #         self._shared_state["not_running_stream"].set() # stop the generator
    #         self._thread.join() # wait for the thread to finish

    #     # clear the data queue
    #     while not self._shared_state["data_queue"].empty(): 
    #         self._shared_state["data_queue"].get()

    #     # start the new stream
    #     prompt_end = prompt.find(b'<|endofprompt|>')
    #     if prompt_end >= 0:
    #         stripped_prompt = prompt[:prompt_end]
    #     else:
    #         raise Exception("This model cannot handle prompts that don't match the instruct format!")
    #     self._shared_state["not_running_stream"].clear() # so we know we are running
    #     # self._stop_event.clear()
    #     self._shared_state["data"] = stripped_prompt + b'<|endofprompt|>'# we start with this data
    #     self._shared_state["num_calls_made"] += 1
    #     generator = self.model_obj.predict_streaming(prompt.decode("utf8"), max_output_tokens=self.max_streaming_tokens)
    #     self._thread = threading.Thread(target=self._start_generator_stream, args=(generator,))
    #     self._thread.start()
    #     # self._start_generator_stream(stripped_prompt)

    # def _start_new_stream(self, prompt):

    #     if self._shared_state["num_calls_made"] > self.max_calls:
    #         raise Exception(f"We have exceeded the maximum number of calls! {self.max_calls}")


    #     # stop any running stream
    #     if self._running_stream():
    #         self._shared_state["not_running_stream"].set() # stop the generator
    #         self._thread.join() # wait for the thread to finish

    #     # clear the data queue
    #     while not self._shared_state["data_queue"].empty(): 
    #         self._shared_state["data_queue"].get()

    #     # start the new stream
    #     generator = self._generator(prompt)
    #     self._shared_state["not_running_stream"].clear() # so we know we are running
    #     self._shared_state["num_calls_made"] += 1

    #     self._thread = threading.Thread(target=self._start_generator_stream, args=(generator,))
    #     self._thread.start()
    #     # self._start_generator_stream(stripped_prompt)
    
    # def _get_logits(self, token_ids, forced_bytes):
    #     '''Computes the logits for the given token state.
        
    #     This overrides a method from the Local class that is used to get
    #     inference results from the model.
    #     '''

    #     if len(token_ids) == 0:
    #         raise ValueError("token_ids must contain some tokens.")
        
    #     # compute the prompt bytes
    #     prompt = b''.join([self.tokens[i] for i in token_ids]) + forced_bytes

    #     self._shared_state["last_call"] = time.time()

    #     # keep looping until we have at least one more byte past our prompt
    #     token_id = None
    #     while True:

    #         # try and get the next token id
    #         if self._shared_state["data"].startswith(prompt):
    #             token_id = self._get_next_token(len(prompt)-len(forced_bytes))
    #             if token_id is not None:
    #                 break
    #         # if self._shared_state["data"].startswith(prompt):
    #         #     token_id = self._get_next_token(len(prompt))#or len(self._shared_state["data"]) <= len(prompt):
    #         #     continue

    #         # we need to restart if extending our data will never lead to matching our prompt
    #         if len(self._shared_state["data"]) <= len(prompt) and not prompt.startswith(self._shared_state["data"]):
    #             self._start_new_stream(prompt)

    #         # extend our data with a chunk from the model stream
    #         if not self._shared_state["data_queue"].empty():
    #             new_bytes = self._shared_state["data_queue"].get_nowait()
                
    #             # if we are at the end of the generation then we try again allowing for early token stopping
    #             if len(new_bytes) == 0:
    #                 token_id = self._get_next_token(len(prompt), allow_early_stop=True)
    #                 if token_id is not None:
    #                     break
    #             self._shared_state["data"] += new_bytes
            
    #         # but if there is nothing and we are not running then we start a stream
    #         elif self._shared_state["not_running_stream"].is_set():
    #             self._start_new_stream(prompt)

    #         # we wait for the running stream to put something in the queue
    #         else:
    #             self._shared_state["data"] += self._shared_state["data_queue"].get()
        
    #     # # if we don't have the next byte of data yet then we wait for it (from the streaming thread)
    #     # if len(self._shared_state["data"]) == len(prompt):
    #     #     self._shared_state["data"] += self._shared_state["data"]_queue.get() 

    #     # token_id = self._get_next_token(len(prompt))

    #     # set the logits to the next byte the model picked
    #     logits = np.ones(len(self.tokens)) * np.nan
    #     logits[token_id] = 100
        
    #     self._shared_state["last_call"] = time.time()
    #     return torch.tensor(logits) # TODO: the caller need to know to NOT use the 0 entries, but to fail out if that happens
    
    # def _get_next_token(self, pos, allow_early_stop=False):
    #     data = self._shared_state["data"]
    #     trie = self._token_trie
    #     token_id = None
    #     while True:
            
    #         # see if we have run out of data
    #         if pos >= len(data):
    #             if allow_early_stop:
    #                 return token_id
    #             else:
    #                 return None
            
    #         # try and walk down the trie
    #         next_byte = data[pos:pos+1]
    #         if next_byte in trie.children:
    #             trie = trie.children[next_byte]
    #             pos += 1
    #             if trie.value is not None:
    #                 token_id = trie.value
    #         else:
    #             return token_id # this is the longest greedy token match we can make

    # def _call_end(self):
    #     if 
    #     self._stop_event.set() # stop the generator
    #     self._generator = None
    #     self._thread.join() # stop the streaming thread

class VertexAIInstruct(VertexAI, Instruct):

    def get_role_start(self, name):
        return ""
    
    def get_role_end(self, name):
        if name == "instruction":
            return "<|endofprompt|>"
        else:
            raise Exception(f"The VertexAIInstruct model does not know about the {name} role type!")

    def _generator(self, prompt):
        # start the new stream
        prompt_end = prompt.find(b'<|endofprompt|>')
        if prompt_end >= 0:
            stripped_prompt = prompt[:prompt_end]
        else:
            raise Exception("This model cannot handle prompts that don't match the instruct format!")
        self._shared_state["not_running_stream"].clear() # so we know we are running
        self._shared_state["data"] = stripped_prompt + b'<|endofprompt|>'# we start with this data
        for chunk in self.model_obj.predict_streaming(self._shared_state["data"].decode("utf8"), max_output_tokens=self.max_streaming_tokens):
            yield chunk.text.encode("utf8")
    
    # def _get_next_token(self, pos, allow_early_stop=False):
    #     data = self._shared_state["data"]
    #     trie = self._token_trie
    #     token_id = None
    #     while True:
            
    #         # see if we have run out of data
    #         if pos >= len(data):
    #             if allow_early_stop:
    #                 return token_id
    #             else:
    #                 return None
            
    #         # try and walk down the trie
    #         next_byte = data[pos:pos+1]
    #         if next_byte in trie.children:
    #             trie = trie.children[next_byte]
    #             pos += 1
    #             if trie.value is not None:
    #                 token_id = trie.value
    #         else:
    #             return token_id # this is the longest greedy token match we can make

    # def _call_end(self):
    #     if 
    #     self._stop_event.set() # stop the generator
    #     self._generator = None
    #     self._thread.join() # stop the streaming thread

class VertexAIChat(VertexAI, Chat):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _generator(self, prompt):
        
        # find the system text
        pos = 0
        system_start = b'<|im_start|>system\n'
        user_start = b'<|im_start|>user\n'
        assistant_start = b'<|im_start|>assistant\n'
        role_end = b'<|im_end|>'
        # system_start_pos = prompt.startswith(system_start)
        
        # find the system text
        system_text = b''
        if prompt.startswith(system_start):
            pos += len(system_start)
            system_end_pos = prompt.find(role_end)
            system_text = prompt[pos:system_end_pos]
            pos = system_end_pos + len(role_end)

        # find the user/assistant pairs
        user_assistant_pairs = []
        last_user_text = b''
        while True:

            # find the user text
            if prompt[pos:].startswith(user_start):
                pos += len(user_start)
                user_end_pos = prompt[pos:].find(role_end)
                user_text = prompt[pos:pos+user_end_pos]
                pos += user_end_pos + len(role_end)
            else:
                raise Exception(f"Bad chat format! Expected a user start tag where there was none at prompt byte position: {pos}")
            
            # and the assistant text
            if prompt[pos:].startswith(assistant_start):
                pos += len(assistant_start)
                assistant_end_pos = prompt[pos:].find(role_end)
                if assistant_end_pos < 0:
                    last_user_text = user_text
                    break
                else:
                    assistant_text = prompt[pos:pos+assistant_end_pos]
                    pos += assistant_end_pos + len(role_end)
                    user_assistant_pairs.append((user_text, assistant_text))
            else:
                raise Exception(f"Bad chat format! Expected a assistant start tag where there was none at prompt byte position: {pos}")
            
        self._shared_state["data"] = prompt[:pos]
        
        chat = self.model_obj.start_chat(
            context=system_text.decode("utf8"),
            examples=[
                vertexai.language_models.InputOutputTextPair(
                    input_text=pair[0].decode("utf8"),
                    output_text=pair[1].decode("utf8"),
                )
            for pair in user_assistant_pairs],
        )

        generator = chat.send_message_streaming(
            last_user_text.decode("utf8"),
            max_output_tokens=self.max_streaming_tokens
        )

        for chunk in generator:
            yield chunk.text.encode("utf8")


import guidance
from guidance import select, models, gen
from ..utils import get_model

def test_fstring():
    lm = get_model("transformers:gpt2")
    lm += f'this is a test {select(["item1", "item2"])}'
    assert str(lm) in ["this is a test item1", "this is a test item2"]

def test_fstring_custom():
    lm = get_model("transformers:gpt2")

    @guidance
    def my_function(lm):
        return lm + f'another {select(["item1", "item2"])}'

    lm += f'this is a test {my_function()}'
    assert str(lm) in ["this is a test another item1", "this is a test another item2"]

def test_token_count():
    lm = get_model("transformers:gpt2")
    lm2 = lm + ' 1 1 1 1 1' + gen(max_tokens=9) + gen(max_tokens=9)
    assert 18 <= lm2._token_count <= 20 # note we allow ourselves to be off by one because it is hard to know when we are continuing vs starting a new token in the parser

def test_call_embeddings():
    '''This tests calls embedded in strings.'''
    model = models.LocalMock()

    @guidance(dedent=False)
    def bla(lm, bla):
        lm += bla + 'ae' + gen(max_tokens=10)
        return lm
    
    @guidance(dedent=False)
    def ble(lm):
        lm += f'''
    ae galera! {bla('33')}
    let's do more stuff!!''' + gen(max_tokens=10)
        return lm
    
    assert "{{G|" not in str(model + ble())


import openai
import os
import time
import requests
import aiohttp
import copy
import time
import asyncio
import types
import collections
import json
import re
import regex
import pyparsing as pp

from ._llm import LLM, LLMSession, SyncSession


class OpenAI(LLM):
    llm_name: str = "openai"

    def __init__(self, model=None, caching=True, max_retries=5, max_calls_per_min=60,
                 api_key=None, api_type="open_ai", api_base=None, api_version=None, deployment_id=None,
                 temperature=0.0, chat_mode="auto", organization=None, rest_call=False,
                 allowed_special_tokens={"<|endoftext|>", "<|endofprompt|>"},
                 token=None, endpoint=None):
        super().__init__()

        # map old param values
        # TODO: add deprecated warnings after some time
        if token is not None:    
            if api_key is None:
                api_key = token
        if endpoint is not None:
            if api_base is None:
                api_base = endpoint

        # fill in default model value
        if model is None:
            model = os.environ.get("OPENAI_MODEL", None)
        if model is None:
            try:
                with open(os.path.expanduser('~/.openai_model'), 'r') as file:
                    model = file.read().replace('\n', '')
            except:
                pass

        # fill in default deployment_id value
        if deployment_id is None:
            deployment_id = os.environ.get("OPENAI_DEPLOYMENT_ID", None)

        # auto detect chat completion mode
        if chat_mode == "auto":
            # parse to determin if the model need to use the chat completion API
            chat_model_pattern = r'^(gpt-3\.5-turbo|gpt-4)(-\d+k)?(-\d{4})?$'
            if re.match(chat_model_pattern, model):
                chat_mode = True
            else:
                chat_mode = False
        
        # fill in default API key value
        if api_key is None: # get from environment variable
            api_key = os.environ.get("OPENAI_API_KEY", getattr(openai, "api_key", None))
        if api_key is not None and not api_key.startswith("sk-") and os.path.exists(os.path.expanduser(api_key)): # get from file
            with open(os.path.expanduser(api_key), 'r') as file:
                api_key = file.read().replace('\n', '')
        if api_key is None: # get from default file location
            try:
                with open(os.path.expanduser('~/.openai_api_key'), 'r') as file:
                    api_key = file.read().replace('\n', '')
            except:
                pass
        if organization is None:
            organization = os.environ.get("OPENAI_ORGANIZATION", None)
        # fill in default endpoint value
        if api_base is None:
            api_base = os.environ.get("OPENAI_API_BASE", None) or os.environ.get("OPENAI_ENDPOINT", None) # ENDPOINT is deprecated

        import tiktoken
        self._tokenizer = tiktoken.get_encoding(tiktoken.encoding_for_model(model).name)
        self.chat_mode = chat_mode
        
        self.allowed_special_tokens = allowed_special_tokens
        self.model_name = model
        self.deployment_id = deployment_id
        self.caching = caching
        self.max_retries = max_retries
        self.max_calls_per_min = max_calls_per_min
        if isinstance(api_key, str):
            api_key = api_key.replace("Bearer ", "")
        self.api_key = api_key
        self.api_type = api_type
        self.api_base = api_base
        self.api_version = api_version
        self.current_time = time.time()
        self.call_history = collections.deque()
        self.temperature = temperature
        self.organization = organization
        self.rest_call = rest_call
        self.endpoint = endpoint

        if not self.rest_call:
            self.caller = self._library_call
        else:
            self.caller = self._rest_call
            self._rest_headers = {
                "Content-Type": "application/json"
            }

    def session(self, asynchronous=False):
        if asynchronous:
            return OpenAISession(self)
        else:
            return SyncSession(OpenAISession(self))

    def role_start(self, role_name, **kwargs):
        assert self.chat_mode, "role_start() can only be used in chat mode"
        return "<|im_start|>"+role_name+"".join([f' {k}="{v}"' for k,v in kwargs.items()])+"\n"
    
    def role_end(self, role=None):
        assert self.chat_mode, "role_end() can only be used in chat mode"
        return "<|im_end|>"
    
    def end_of_text(self):
        return "<|endoftext|>"
    
    @classmethod
    async def stream_then_save(cls, gen, key, stop_regex, n):
        list_out = []
        cached_out = None

        # init stop_regex variables
        if stop_regex is not None:
            if isinstance(stop_regex, str):
                stop_patterns = [regex.compile(stop_regex)]
            else:
                stop_patterns = [regex.compile(pattern) for pattern in stop_regex]

            current_strings = ["" for _ in range(n)]
            # last_out_pos = ["" for _ in range(n)]
        
        # iterate through the stream
        all_done = False
        async for curr_out in gen:

            # if we have a cached output, extend it with the current output
            if cached_out is not None:
                out = merge_stream_chunks(cached_out, curr_out)
            else:
                out = curr_out
            
            # check if we have stop_regex matches
            found_partial = False
            if stop_regex is not None:

                # keep track of the generated text so far
                for i,choice in enumerate(curr_out['choices']):
                    current_strings[i] += choice['text']

                # check if all of the strings match a stop string (and hence we can stop the batch inference)
                all_done = True
                for i in range(len(current_strings)):
                    found = False
                    for s in stop_patterns:
                        if s.search(current_strings[i]):
                            found = True
                    if not found:
                        all_done = False
                        break

                # find where trim off the stop regex matches if needed (and look for partial matches)
                stop_pos = [1e10 for _ in range(n)]
                stop_text = [None for _ in range(n)]
                for i in range(len(current_strings)):
                    for s in stop_patterns:
                        m = s.search(current_strings[i], partial=True)
                        if m:
                            span = m.span()
                            if span[1] > span[0]:
                                if m.partial: # we might be starting a stop sequence, so we can't emit anything yet
                                    found_partial = True
                                    break
                                else:
                                    stop_text[i] = current_strings[i][span[0]:span[1]]
                                    stop_pos[i] = min(span[0], stop_pos[i])
                    if stop_pos != 1e10:
                        stop_pos[i] = stop_pos[i] - len(current_strings[i]) # convert to relative position from the end
            
            # if we might be starting a stop sequence, we need to cache the output and continue to wait and see
            if found_partial:
                cached_out = out
                continue
            
            # if we get here, we are not starting a stop sequence, so we can emit the output
            else:
                cached_out = None

                if stop_regex is not None:
                    for i in range(len(out['choices'])):
                        if stop_pos[i] < len(out['choices'][i]['text']):
                            out['choices'][i] = out['choices'][i].to_dict() # because sometimes we might need to set the text to the empty string (and OpenAI's object does not like that)
                            out['choices'][i]['text'] = out['choices'][i]['text'][:stop_pos[i]]
                            out['choices'][i]['stop_text'] = stop_text[i]
                            out['choices'][i]['finish_reason'] = "stop"
            
                list_out.append(out)
                yield out
                if all_done:
                    gen.aclose()
                    break
        
        # if we have a cached output, emit it
        if cached_out is not None:
            list_out.append(cached_out)
            yield out

        cls.cache[key] = list_out
    
    def _stream_completion(self):
        pass

    # Define a function to add a call to the deque
    def add_call(self):
        # Get the current timestamp in seconds
        now = time.time()
        # Append the timestamp to the right of the deque
        self.call_history.append(now)

    # Define a function to count the calls in the last 60 seconds
    def count_calls(self):
        # Get the current timestamp in seconds
        now = time.time()
        # Remove the timestamps that are older than 60 seconds from the left of the deque
        while self.call_history and self.call_history[0] < now - 60:
            self.call_history.popleft()
        # Return the length of the deque as the number of calls
        return len(self.call_history)

    async def _library_call(self, **kwargs):
        """ Call the OpenAI API using the python package.

        Note that is uses the local auth token, and does not rely on the openai one.
        """

        # save the params of the openai library
        prev_key = openai.api_key
        prev_org = openai.organization
        prev_type = openai.api_type
        prev_version = openai.api_version
        prev_base = openai.api_base
        
        # set the params of the openai library if we have them
        if self.api_key is not None:
            openai.api_key = self.api_key
        if self.organization is not None:
            openai.organization = self.organization
        if self.api_type is not None:
            openai.api_type = self.api_type
        if self.api_version is not None:
            openai.api_version = self.api_version
        if self.api_base is not None:
            openai.api_base = self.api_base

        assert openai.api_key is not None, "You must provide an OpenAI API key to use the OpenAI LLM. Either pass it in the constructor, set the OPENAI_API_KEY environment variable, or create the file ~/.openai_api_key with your key in it."
        
        if self.chat_mode:
            kwargs['messages'] = prompt_to_messages(kwargs['prompt'])
            del kwargs['prompt']
            del kwargs['echo']
            del kwargs['logprobs']
            # print(kwargs)
            out = await openai.ChatCompletion.acreate(**kwargs)
            out = add_text_to_chat_mode(out)
        else:
            out = await openai.Completion.acreate(**kwargs)
        
        # restore the params of the openai library
        openai.api_key = prev_key
        openai.organization = prev_org
        openai.api_type = prev_type
        openai.api_version = prev_version
        openai.api_base = prev_base
        
        return out

    async def _rest_call(self, **kwargs):
        """ Call the OpenAI API using the REST API.
        """

        # Define the request headers
        headers = copy.copy(self._rest_headers)
        if self.api_key is not None:
            headers['Authorization'] = f"Bearer {self.api_key}"

        # Define the request data
        stream = kwargs.get("stream", False)
        data = {
            "model": self.model_name,
            "prompt": kwargs["prompt"],
            "max_tokens": kwargs.get("max_tokens", None),
            "temperature": kwargs.get("temperature", 0.0),
            "top_p": kwargs.get("top_p", 1.0),
            "n": kwargs.get("n", 1),
            "stream": stream,
            "logprobs": kwargs.get("logprobs", None),
            'stop': kwargs.get("stop", None),
            "echo": kwargs.get("echo", False)
        }
        if self.chat_mode:
            data['messages'] = prompt_to_messages(data['prompt'])
            del data['prompt']
            del data['echo']
            del data['logprobs']

        # Send a POST request and get the response
        # An exception for timeout is raised if the server has not issued a response for 10 seconds
        try:
            if stream:
                session = aiohttp.ClientSession()
                response = await session.post(self.endpoint, json=data, headers=headers, timeout=60)
                status = response.status
            else:
                response = requests.post(self.endpoint, headers=headers, json=data, timeout=60)
                status = response.status_code
                text = response.text
            if status != 200:
                if stream:
                    text = await response.text()
                raise Exception("Response is not 200: " + text)
            if stream:
                response = self._rest_stream_handler(response, session)
            else:
                response = response.json()
        except requests.Timeout:
            raise Exception("Request timed out.")
        except requests.ConnectionError:
            raise Exception("Connection error occurred.")
        if self.chat_mode:
            response = add_text_to_chat_mode(response)
        return response
        
    async def _close_response_and_session(self, response, session):
        await response.release()
        await session.close()

    async def _rest_stream_handler(self, response, session):
        # async for line in response.iter_lines():
        async for line in response.content:
            text = line.decode('utf-8')
            if text.startswith('data: '):
                text = text[6:]
                if text.strip() == '[DONE]':
                    await self._close_response_and_session(response, session)
                    break
                else:
                    yield json.loads(text)
    
    def encode(self, string):
        # note that is_fragment is not used used for this tokenizer
        return self._tokenizer.encode(string, allowed_special=self.allowed_special_tokens)
    
    def decode(self, tokens):
        return self._tokenizer.decode(tokens)

def add_text_to_chat_mode(chat_mode):
    """Adds the text field to the chat mode response, so that it is consistent with the non-chat mode response."""
    if isinstance(chat_mode, (types.AsyncGeneratorType, types.GeneratorType)):
        return add_text_to_chat_mode_generator(chat_mode)
    else:
        for c in chat_mode['choices']:
            c['text'] = c['message']['content']
        return chat_mode

async def add_text_to_chat_mode_generator(chat_mode):
    """Same as add_text_to_chat_mode, but for async generators."""
    in_function_call = False
    async for resp in chat_mode:
        if "choices" in resp:
            for c in resp['choices']:
                
                # move content from delta to text so we have a consistent interface with non-chat mode
                found_content = False
                if "content" in c['delta'] and c['delta']['content'] != "":
                    found_content = True
                    c['text'] = c['delta']['content']

                # capture function call data and convert to text again so we have a consistent interface with non-chat mode and open models
                if "function_call" in c['delta']:

                    # build the start of the function call (the follows the syntax that GPT says it wants when we ask it, and will be parsed by the @function_detector)
                    if not in_function_call:
                        start_val = "\n```typescript\nfunctions."+c['delta']['function_call']["name"]+"("
                        if not c['text']:
                            c['text'] = start_val
                        else:
                            c['text'] += start_val
                        in_function_call = True
                    
                    # extend the arguments JSON string
                    val = c['delta']['function_call']["arguments"]
                    if 'text' in c:
                        c['text'] += val
                    else:
                        c['text'] = val
                    
                if not found_content and not in_function_call:
                    break # the role markers are outside the generation in chat mode right now TODO: consider how this changes for uncontrained generation
            else:
                yield resp
        else:
            yield resp
    
    # close the function call if needed
    if in_function_call:
        yield {'choices': [{'text': ')```'}]}

def merge_stream_chunks(first_chunk, second_chunk):
    """ This merges two stream responses together.
    """

    out = copy.deepcopy(first_chunk)

    # merge the choices
    for i in range(len(out['choices'])):
        out_choice = out['choices'][i]
        second_choice = second_chunk['choices'][i]
        out_choice['text'] += second_choice['text']
        if 'index' in second_choice:
            out_choice['index'] = second_choice['index']
        if 'finish_reason' in second_choice:
            out_choice['finish_reason'] = second_choice['finish_reason']
        if out_choice.get('logprobs', None) is not None:
            out_choice['logprobs']['token_logprobs'] += second_choice['logprobs']['token_logprobs']
            out_choice['logprobs']['top_logprobs'] += second_choice['logprobs']['top_logprobs']
            out_choice['logprobs']['text_offset'] = second_choice['logprobs']['text_offset']
    
    return out
    
# define a grammar for parsing roles
role_start_tag = pp.Suppress(pp.Optional(pp.White()) + pp.Literal("<|im_start|>"))
role_start_name = pp.Word(pp.alphanums + "_")("role_name")
role_kwargs = pp.Suppress(pp.Optional(" ")) + pp.Dict(pp.Group(pp.Word(pp.alphanums + "_") + pp.Suppress("=") + pp.QuotedString('"')))("kwargs")
role_start = (role_start_tag + role_start_name + pp.Optional(role_kwargs) + pp.Suppress("\n")).leave_whitespace()
role_end = pp.Suppress(pp.Literal("<|im_end|>"))
role_content = pp.Combine(pp.ZeroOrMore(pp.CharsNotIn("<") | pp.Literal("<") + ~pp.FollowedBy("|im_end|>")))("role_content")
role_group = pp.Group(role_start + role_content + role_end)("role_group").leave_whitespace()
partial_role_group = pp.Group(role_start + role_content)("role_group").leave_whitespace()
roles_grammar = pp.ZeroOrMore(role_group) + pp.Optional(partial_role_group) + pp.StringEnd()

def prompt_to_messages(prompt):
    """Takes a raw prompt string and returns a list of messages with role and content fields for API calls."""
    messages = []

    # TODO: remove this in case some endpoints support partial prompting?
    assert prompt.endswith("<|im_start|>assistant\n"), "When calling OpenAI chat models you must generate only directly inside the assistant role! The OpenAI API does not currently support partial assistant prompting."

    # use pyparsing to parse the prompt into roles
    parsed_prompt = roles_grammar.parse_string(prompt)

    # convert the parsed prompt into a list of messages for the API
    for role in parsed_prompt:
        if len(role["role_content"]) > 0: # only add non-empty messages (OpenAI does not support empty messages anyway)
            message = {'role': role["role_name"], 'content': role["role_content"]}
            if "kwargs" in role:
                for k, v in role["kwargs"].items():
                    message[k] = v
            messages.append(message)

    return messages

# define a grammar for parsing function definitions
import pyparsing as pp
start_functions = pp.Suppress(pp.Literal("## functions\n\nnamespace functions {\n\n"))
comment = pp.Combine(pp.Suppress(pp.Literal("//") + pp.Optional(" ")) + pp.restOfLine)
end_functions = pp.Suppress("} // namespace functions")
function_def_start = pp.Optional(comment)("function_description") + pp.Suppress(pp.Literal("type")) + pp.Word(pp.alphas + "_")("function_name") + pp.Suppress(pp.Literal("=") + pp.Literal("(_:") + pp.Literal("{"))
function_def_end = pp.Suppress(pp.Literal("})") + pp.Literal("=>") + pp.Literal("any;"))
parameter_type = (pp.Word(pp.alphas + "_")("simple_type") | pp.QuotedString('"')("enum_option") + pp.OneOrMore(pp.Suppress("|") + pp.QuotedString('"')("enum_option"))("enum")) + pp.Suppress(pp.Optional(","))
parameter_def = pp.Optional(comment)("parameter_description") + pp.Word(pp.alphas + "_")("parameter_name") + pp.Optional(pp.Literal("?"))("is_optional") + pp.Suppress(pp.Literal(":")) + pp.Group(parameter_type)("parameter_type")
function_def = function_def_start + pp.OneOrMore(pp.Group(parameter_def)("parameter")) + function_def_end
functions_def = start_functions + pp.OneOrMore(pp.Group(function_def)("function")) + end_functions

def extract_function_defs(prompt):
    """ This extracts function definitions from the prompt.
    """

    if "\n## functions\n" not in prompt:
        return None
    else:
        functions_text = prompt[prompt.index("\n## functions\n")+1:prompt.index("} // namespace functions")+24]
        parse_out = functions_def.parse_string(functions_text)
        return get_json_from_parse(parse_out)

def get_json_from_parse(parse_out):
    functions = []
    for function in parse_out:
        function_name = function.function_name
        function_description = function.function_description
        parameters = {
            "type": "object",
            "properties": {},
            "required": []
        }
        for parameter in function:
            if isinstance(parameter, str):
                continue
            parameter_name = parameter.parameter_name
            parameter_description = parameter.parameter_description
            parameter_type = parameter.parameter_type
            is_optional = parameter.is_optional
            d = {}
            if parameter_type.simple_type:
                d["type"] = parameter_type.simple_type
            elif parameter_type.enum:
                d["type"] = "string"
                d["enum"] = [s for s in parameter_type]
            if parameter_description:
                d["description"] = parameter_description
            if not is_optional:
                parameters["required"].append(parameter_name)
            parameters["properties"][parameter_name] = d
        functions.append({
            "name": function_name,
            "description": function_description,
            "parameters": parameters
        })
    return functions

class OpenAISession(LLMSession):
    async def __call__(self, prompt, stop=None, stop_regex=None, temperature=None, n=1, max_tokens=1000, logprobs=None,
                       top_p=1.0, echo=False, logit_bias=None, token_healing=None, pattern=None, stream=None,
                       cache_seed=0, caching=None, **completion_kwargs):
        """ Generate a completion of the given prompt.
        """

        # we need to stream in order to support stop_regex
        if stream is None:
            stream = stop_regex is not None
        assert stop_regex is None or stream, "We can only support stop_regex for the OpenAI API when stream=True!"
        assert stop_regex is None or n == 1, "We don't yet support stop_regex combined with n > 1 with the OpenAI API!"

        assert token_healing is None or token_healing is False, "The OpenAI API does not yet support token healing! Please either switch to an endpoint that does, or don't use the `token_healing` argument to `gen`."

        # set defaults
        if temperature is None:
            temperature = self.llm.temperature

        # get the arguments as dictionary for cache key generation
        args = locals().copy()

        assert not pattern, "The OpenAI API does not support Guidance pattern controls! Please either switch to an endpoint that does, or don't use the `pattern` argument to `gen`."
        # assert not stop_regex, "The OpenAI API does not support Guidance stop_regex controls! Please either switch to an endpoint that does, or don't use the `stop_regex` argument to `gen`."

        # define the key for the cache
        cache_params = self._cache_params(args)
        llm_cache = self.llm.cache
        key = llm_cache.create_key(self.llm.llm_name, **cache_params)
        
        # allow streaming to use non-streaming cache (the reverse is not true)
        if key not in llm_cache and stream:
            cache_params["stream"] = False
            key1 = llm_cache.create_key(self.llm.llm_name, **cache_params)
            if key1 in llm_cache:
                key = key1
        
        # check the cache
        if key not in llm_cache or caching is False or (caching is not True and not self.llm.caching):

            # ensure we don't exceed the rate limit
            while self.llm.count_calls() > self.llm.max_calls_per_min:
                await asyncio.sleep(1)

            functions = extract_function_defs(prompt)

            fail_count = 0
            while True:
                try_again = False
                try:
                    self.llm.add_call()
                    call_args = {
                        "model": self.llm.model_name,
                        "deployment_id": self.llm.deployment_id,
                        "prompt": prompt,
                        "max_tokens": max_tokens,
                        "temperature": temperature,
                        "top_p": top_p,
                        "n": n,
                        "stop": stop,
                        "logprobs": logprobs,
                        "echo": echo,
                        "stream": stream,
                        **completion_kwargs
                    }
                    if functions is None:
                        if "function_call" in call_args:
                            del call_args["function_call"]
                    else:
                        call_args["functions"] = functions
                    if logit_bias is not None:
                        call_args["logit_bias"] = {str(k): v for k,v in logit_bias.items()} # convert keys to strings since that's the open ai api's format
                    out = await self.llm.caller(**call_args)

                except (openai.error.RateLimitError, openai.error.ServiceUnavailableError):
                    await asyncio.sleep(3)
                    try_again = True
                    fail_count += 1
                
                if not try_again:
                    break

                if fail_count > self.llm.max_retries:
                    raise Exception(f"Too many (more than {self.llm.max_retries}) OpenAI API RateLimitError's or ServiceUnavailableError's in a row!")

            if stream:
                return self.llm.stream_then_save(out, key, stop_regex, n)
            else:
                llm_cache[key] = out
        
        # wrap as a list if needed
        if stream:
            if isinstance(llm_cache[key], list):
                return llm_cache[key]
            return [llm_cache[key]]
        
        return llm_cache[key]


import os
import json
import platformdirs
from ._openai import OpenAI

class AzureOpenAI(OpenAI):
    def __init__(self, *args, **kwargs):
        raise NotImplementedError("The AzureOpenAI class has been merged with the OpenAI class for Azure usage. Please use the OpenAI class instead: https://guidance.readthedocs.io/en/latest/example_notebooks/api_examples/llms/OpenAI.html")

class MSALOpenAI(OpenAI):
    """ Microsoft Authentication Library (MSAL) OpenAI style integration.

    Warning: This class is not finalized and may change in the future.
    """

    llm_name: str = "azure_openai"

    def __init__(self, model=None, client_id=None, authority=None, caching=True, max_retries=5, max_calls_per_min=60, token=None,
                 endpoint=None, scopes=None, temperature=0.0, chat_mode="auto"):
        

        assert endpoint is not None, "An endpoint must be specified!"
        
        # build a standard OpenAI LLM object
        super().__init__(
            model=model, caching=caching, max_retries=max_retries, max_calls_per_min=max_calls_per_min,
            token=token, endpoint=endpoint, temperature=temperature, chat_mode=chat_mode
        )

        self.client_id = client_id
        self.authority = authority
        self.scopes = scopes

        from msal import PublicClientApplication, SerializableTokenCache
        self._token_cache = SerializableTokenCache()
        self._token_cache_path = os.path.join(platformdirs.user_cache_dir("guidance"), "_azure_openai.token")
        self._app = PublicClientApplication(client_id=self.client_id, authority=self.authority, token_cache=self._token_cache)
        if os.path.exists(self._token_cache_path):
            self._token_cache.deserialize(open(self._token_cache_path, 'r').read())

        self._rest_headers["X-ModelType"] = self.model_name

    @property
    def token(self):
        return self._get_token()
    @token.setter
    def token(self, value):
        pass # ignored for now

    def _get_token(self):
        accounts = self._app.get_accounts()
        result = None

        if accounts:
            # Assuming the end user chose this one
            chosen = accounts[0]

            # Now let's try to find a token in cache for this account
            result = self._app.acquire_token_silent(self.scopes, account=chosen)
    
        if not result:
            # So no suitable token exists in cache. Let's get a new one from AAD.
            flow = self._app.initiate_device_flow(scopes=self.scopes)

            if "user_code" not in flow:
                raise ValueError(
                    "Fail to create device flow. Err: %s" % json.dumps(flow, indent=4))

            print(flow["message"])

            result = self._app.acquire_token_by_device_flow(flow)

            # save the aquired token
            with open(self._token_cache_path, "w") as f:
                f.write(self._token_cache.serialize())

        return result["access_token"]



__version__ = "0.1.2"

import nest_asyncio
nest_asyncio.apply()
import types
import sys
import os
import requests
from . import models
import inspect

from ._utils import load, chain, Silent, Hidden, CaptureEvents, TextRange, strip_multiline_string_indents
from . import _utils
from . import selectors

from ._grammar import StatelessFunction, StatefulFunction, string, Terminal, Placeholder, replace_grammar_node
import functools
from contextlib import nullcontext

curr_module = sys.modules[__name__]

# the user needs to set an LLM before they can use guidance
llm = None

newline = "\n"

# This makes the guidance module callable
class Guidance(types.ModuleType):
    def __call__(self, f=None, *, stateless=False, cache=None, dedent=True, model=models.Model):
        return _decorator(f, stateless=stateless, cache=cache, dedent=dedent, model=model)
sys.modules[__name__].__class__ = Guidance

def optional_hidden(f, lm, hidden, kwargs):
    """This only enters a hidden context if the function does not manage the hidden parameter itself.
    """
    if 'hidden' in inspect.signature(f).parameters:
        kwargs['hidden'] = hidden
        return nullcontext()
    else:
        return Hidden(lm, hidden)
    
_function_cache = {} # used to enable recursive grammar definitions
_null_grammar = string('')
# TODO: enable streaming for guidance function evaluation

def _decorator(f, *, stateless, cache, dedent, model):
    
    # if we are not yet being used as a decorator, then save the args
    if f is None:
        return functools.partial(_decorator, stateless=stateless, cache=cache, dedent=dedent, model=model)
    
    # if we are being used as a decorator then return the decorated function
    else:

        # this strips out indentation in multiline strings that aligns with the current python indentation
        if dedent is True or dedent == 'python':
            f = strip_multiline_string_indents(f)

        # we cache if requested
        if cache:
            f = functools.cache(f)

        @functools.wraps(f)
        def wrapped(*args, **kwargs):

            # make a stateless grammar if we can
            if stateless is True or (callable(stateless) and stateless(*args, **kwargs)):
                
                # if we have a placeholder set then we must be in a recursive definition and so we return the placeholder
                placeholder = getattr(f, "_self_call_placeholder_", None)
                if placeholder is not None:
                    return placeholder
                
                # otherwise we call the function to generate the grammar
                else:
                    
                    # set a placeholder for recursive calls (only if we don't have arguments that might make caching a bad idea)
                    no_args = len(args) + len(kwargs) == 0
                    if no_args:
                        f._self_call_placeholder_ = Placeholder()

                    # call the function to get the grammar node
                    node = f(_null_grammar, *args, **kwargs)
                    if not isinstance(node, Terminal):
                        node.name = f.__name__

                    # replace all the placeholders with our generated node
                    if no_args:
                        replace_grammar_node(node, f._self_call_placeholder_, node)
                        del f._self_call_placeholder_

                    return node

            # otherwise must be stateful (which means we can't be inside a select() call)
            else:
                return StatefulFunction(f, args, kwargs)
        
        # attach this as a method of the model class (if given)
        if model is not None:
            setattr(model, f.__name__, f)
        
        return wrapped

# def _decorator(f, *, model=None, dedent='python'):

#     def _decorator_inner(f, model=models.Model):
#         """Decorator to turn a normal function into a guidance function.
        
#         Guidance functions have the added ability to be called as methods of LM objects (for dot-chaining),
#         and to be optionally iterated over to get a stream of results (syncronously or asyncronously).
#         TODO: In the future we plan to add network aware guidance acceleration as well.
#         """

#         # this strips out indentation in multiline strings that aligns with the current python indentation
#         if dedent == 'python':
#             f = strip_multiline_string_indents(f)
        
#         def sync_wrapper(lm, *args, silent=None, hidden=False, **kwargs):
#             with Silent(lm, silent), optional_hidden(f, lm, hidden, kwargs):
#                 return f(lm, *args, **kwargs)

#         def sync_iter_wrapper(lm, *args, silent=None, hidden=False, **kwargs):

#             # create a worker thread and run the function in it
#             with Silent(lm, silent), optional_hidden(f, lm, hidden, kwargs):
#                 with CaptureEvents(lm) as events:
#                     worker_thread = threading.Thread(target=f, args=(lm, *args), kwargs=kwargs)
#                     worker_thread.start()
                
#                     # loop over the queue and display the results
#                     while True:
#                         try:
#                             val = events.get(timeout=0.1)
#                             yield val
#                         except queue.Empty:
#                             if not worker_thread.is_alive():
#                                 break

#         async def async_wrapper(lm, *args, silent=None, hidden=False, **kwargs):
#             with Silent(lm, silent), optional_hidden(f, lm, hidden, kwargs):
#                 return await f(lm, *args, **kwargs)

#         async def async_iter_wrapper(lm, *args, **kwargs):
#             iterator = _utils.ThreadSafeAsyncIterator(sync_iter_wrapper(lm, *args, **kwargs))
#             async for item in iterator:
#                 yield item

#         @functools.wraps(f)
#         def wrapper(*args, stream=False, async_mode=False, **kwargs):

#             # check if we are making a delayed call
#             if len(args) == 0 or not isinstance(args[0], models.Model):
#                 return wrapper.wrapper_delayed(*args, stream=stream, async_mode=async_mode, **kwargs)
#             else:
#                 lm = args[0]
#                 args = args[1:]

#             # if not we execute
#             if async_mode:
#                 if stream:
#                     return async_iter_wrapper(lm, *args, **kwargs)
#                 else:
#                     return async_wrapper(lm, *args, **kwargs)
#             else:
#                 if stream:
#                     return sync_iter_wrapper(lm, *args, **kwargs)
#                 else:
#                     return sync_wrapper(lm, *args, **kwargs)
                    
#         @functools.wraps(f)
#         def wrapper_delayed(*args, stream=False, async_mode=False, **kwargs):
#             '''Converts a guidance function call to a string, so it can be called later once it is added to an LM object.'''
            
#             # save the call in our call pool, ready to be run when it is attached to an LM object
#             id = str(uuid.uuid4())
#             models.Model._call_pool[id] = lambda lm: wrapper(lm, *args, stream=stream, async_mode=async_mode, **kwargs)

#             # return a string representation of this call so it can be combined with other strings/calls
#             return models.Model.tag_start + id + models.Model.tag_end
#         wrapper.wrapper_delayed = wrapper_delayed

#         setattr(model, f.__name__, wrapper) # as a method on the LM object
#         setattr(curr_module, f.__name__, wrapper_delayed) # as a top level class

#         return wrapper

#     if model is None:
#         return _decorator_inner(f)
#     else:
#         return functools.partial(_decorator_inner, model=model)

def load(guidance_file):
    ''' Load a guidance program from the given text file.

    If the passed file is a valid local file it will be loaded directly.
    Otherwise, if it starts with "http://" or "https://" it will be loaded
    from the web.
    '''

    if os.path.exists(guidance_file):
        with open(guidance_file, 'r') as f:
            template = f.read()
    elif guidance_file.startswith('http://') or guidance_file.startswith('https://'):
        template = requests.get(guidance_file).text
    else:
        raise ValueError('Invalid guidance file: %s' % guidance_file)
    
    return sys.modules[__name__](template)

# we expose all the library functions at the top level of the module
from .library import *


import os
from pathlib import Path
import multiprocessing
from itertools import takewhile
import operator

import numpy as np

from ._model import Chat
from ._local import Local
from .._utils import normalize_notebook_stdout_stderr

try:
    # TODO: can we eliminate the torch requirement for llama.cpp by using numpy in the caller instead?
    import torch
    is_torch = True
except ImportError:
    is_torch = False

try:
    import llama_cpp
    is_llama_cpp = True
except ImportError:
    is_llama_cpp = False

class LlamaCpp(Local):
    def __init__(self, model=None, tokenizer=None, echo=True, caching=True, temperature=0.0, **kwargs):
        if not is_torch:
            raise Exception("Please install PyTorch in order to use guidance.models.LlamaCpp!")

        if not is_llama_cpp:
            raise Exception("Please install llama-cpp-python with `pip install llama-cpp-python` in order to use guidance.models.LlamaCpp!")

        if isinstance(model, Path):
            model = str(model)
        if model is None or isinstance(model, str) and len(model.strip()) == 0:
            model = os.environ.get("LLAMA_CPP_MODEL", "")
            if len(model.strip()) == 0:
                try:
                    with open(os.path.expanduser('~/.llama_cpp_model'), 'r') as file:
                        model = file.read().replace('\n', '')
                except:
                    pass
                if len(model.strip()) == 0:
                    raise ValueError("If model is None then a model file must be specified in either the LLAMA_CPP_MODEL environment variable or in the ~/.llama_cpp_model file.")

        if isinstance(model, str):
            self.model = model
            if "n_threads" not in kwargs:
                kwargs["n_threads"] = multiprocessing.cpu_count()
            if "verbose" not in kwargs:
                kwargs["verbose"] = False
            if "n_batch" not in kwargs and "n_ctx" in kwargs:
                kwargs["n_batch"] = kwargs["n_ctx"]

            with normalize_notebook_stdout_stderr():
                self.model_obj = llama_cpp.Llama(model_path=model, **kwargs)
        elif isinstance(model, llama_cpp.Llama):
            self.model = model.__class__.__name__
            self.model_obj = model
        else:
            raise TypeError("model must be None, a file path string, or a llama_cpp.Llama object.")

        if tokenizer is None:
            tokenizer = llama_cpp.LlamaTokenizer(self.model_obj)
        elif not isinstance(tokenizer, llama_cpp.LlamaTokenizer):
            raise TypeError("tokenizer must be None or a llama_cpp.LlamaTokenizer object.")
        self._orig_tokenizer = tokenizer

        #self._n_threads = multiprocessing.cpu_count()
        self._n_vocab = tokenizer.llama.n_vocab()
        self.caching = caching
        self.temperature = temperature

        tokens = [tokenizer.llama.detokenize([i]) for i in range(self._n_vocab)] # note that detokenize returns bytes directly
        tokens[1] = b"<s>" # these are not decoded correctly by llama_cpp
        tokens[2] = b"</s>"
        self.bos_token = "<s>"
        self.eos_token = "</s>"
        super().__init__(
            tokens,
            tokenizer.llama.token_bos(),
            tokenizer.llama.token_eos(),
            echo=echo
        )

        self._cache_state["cache_token_ids"] = []

    def _get_logits(self, token_ids, forced_bytes):
        '''Computes the logits for the given token state.
        
        This overrides a method from the LocalEngine class that is used to get
        inference results from the model.
        '''

        if len(token_ids) == 0:
            raise ValueError("token_ids must contain some tokens.")

        # check what we have already cached
        cache_token_ids = self._cache_state["cache_token_ids"]
        num_cached = sum(takewhile(operator.truth, map(operator.eq, token_ids, cache_token_ids)))
        if num_cached == len(token_ids):
            if num_cached == len(cache_token_ids):
                return self._cache_state["logits"]
            num_cached = num_cached - 1 # llama_cpp doesn't like it when we pass in 0 new tokens, so re-input one
        
        # make sure we don't run off the end of the model's context
        if self.model_obj.n_ctx() < len(token_ids) + 1:
            raise Exception(f"Attempted to use a context length of {len(token_ids)} tokens, but this LlamaCpp model is only configured to support up to {self.model_obj.n_ctx()}!")

        # eval the model
        self._cache_state["cache_token_ids"] = token_ids.copy()
        token_ids = np.array(token_ids[num_cached:], dtype=np.int32)
        token_ids = (llama_cpp.llama_token * len(token_ids))(*token_ids)
        llama_cpp.llama_eval(self.model_obj.ctx, token_ids, len(token_ids), num_cached)#, self._n_threads)
        
        # get the logits
        logits = llama_cpp.llama_get_logits(self.model_obj.ctx)
        logits = np.ctypeslib.as_array(logits, shape=(self._n_vocab,))
        logits = torch.from_numpy(logits)
        self._cache_state["logits"] = logits
        
        return logits
    
class LlamaCppChat(LlamaCpp, Chat):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_role_start(self, role_name, **kwargs):
        if role_name == "user":

            # if we follow an auto-nested system role then we are done
            if self._current_prompt().endswith("\n<</SYS>>\n\n"):
                return ""
            else:
                return "[INST] "
        
        elif role_name == "assistant":
            return " "
        
        elif role_name == "system":
            
            # check if we are already embedded at the top of a user role
            if self._current_prompt().endswith("[INST] "):
                return "<<SYS>>\n"

            # if not then we auto nest ourselves
            else:
                return "[INST] <<SYS>>\n"
    
    def get_role_end(self, role_name=None):
        if role_name == "user":
            return " [/INST]"
        elif role_name == "assistant":
            return " "
        elif role_name == "system":
            return "\n<</SYS>>\n\n"



import guidance
from guidance import zero_or_more, byte_range
from ..utils import get_model

def test_token_healing():
    '''Tests a bug where the space is incorrectly forced as token 220, while it should be not forced it might be extended'''
    gpt2 = get_model("transformers:gpt2")
    lm = gpt2 + ("This is a story of 10 or 5 or " + zero_or_more(byte_range(b'0', b'9')))
    assert len(lm) > len("This is a story of 10 or 5 or ")


from guidance import select, capture, models, one_or_more
from ..utils import get_model

def test_capture():
    model = models.LocalMock()
    model += 'This is' + capture(select(options=['bad', 'quite bad']), name="my_var")
    assert model["my_var"] in ["bad", "quite bad"]

def test_capture_star():
    lm = models.LocalMock(b"<s>1234233234<s>")
    grammar = capture(one_or_more(select(['1', '2'])), name='test')
    lm2 = lm + grammar
    assert lm2['test'] == '12'


from guidance import gen, models, silent
from ..utils import get_model

def test_basic():
    lm = models.LocalMock()
    lm += "Start text"
    with silent():
        lm += "silent text"
    lm += "end text"
    assert "display: none" in lm._state


from guidance import models, one_or_more, char_set

def test_string():
    model = models.LocalMock("<s>aaabc")
    assert str(model + '<s>' + one_or_more("a")) == "<s>aaa"

def test_grammar():
    model = models.LocalMock("<s>bac")
    assert str(model + '<s>' + one_or_more(char_set("ab"))) == "<s>ba"

def test_at_least_one():
    model = models.LocalMock("<s>cbac")
    assert not str(model + '<s>' + one_or_more(char_set("ab"))).startswith("<s>c")


import re
import functools
import time
import collections
import os

try:
    import torch
except ImportError:
    pass

from .._model import Chat
from .._local import Local


class Transformers(Local):
    def __init__(self, model=None, tokenizer=None, echo=True, caching=True, temperature=0.0, device=None, **kwargs):
        
        # fill in default model value
        if model is None:
            model = os.environ.get("TRANSFORMERS_MODEL", None)
        if model is None:
            try:
                with open(os.path.expanduser('~/.transformers_model'), 'r') as file:
                    model = file.read().replace('\n', '')
            except:
                pass

        self.model_obj, self._orig_tokenizer = self._model_and_tokenizer(model, tokenizer, **kwargs)

        if not isinstance(model, str):
            self.model = model.__class__.__name__
        self.caching = caching
        # self.current_time = time.time()
        # self.call_history = collections.deque()
        self.temperature = temperature
        if device is not None: # set the device if requested
            self.model_obj = self.model_obj.to(device)
        self.device = self.model_obj.device # otherwise note the current device

        # build out the set of byte_string tokens
        tkz = self._orig_tokenizer
        if hasattr(tkz, "byte_decoder"):
            byte_tokens = []
            for i in range(tkz.vocab_size):
                byte_coded = bytes([tkz.byte_decoder[c] for c in tkz.convert_ids_to_tokens(i)])
                byte_tokens.append(byte_coded)
        else:
            byte_tokens = [bytes(tkz.convert_tokens_to_string(['a', tkz.convert_ids_to_tokens(i)])[1:], encoding="utf8") for i in range(tkz.vocab_size)]

        # the superclass does most of the work once we have the tokens
        super().__init__(
            byte_tokens,
            tkz.bos_token_id,
            tkz.eos_token_id,
            echo=echo
        )

        self._cache_state["past_key_values"] = None
        self._cache_state["logits"] = None
        self._cache_state["cache_token_ids"] = []

    def _model_and_tokenizer(self, model, tokenizer, **kwargs):

        # intantiate the model and tokenizer if needed
        if isinstance(model, str):

            # make sure transformers is installed
            try:
                import transformers
            except:
                raise Exception("Please install transformers with `pip install transformers` in order to use guidance.models.Transformers!")

            if tokenizer is None:
                tokenizer = transformers.AutoTokenizer.from_pretrained(model, use_fast=False, **kwargs)
            model = transformers.AutoModelForCausalLM.from_pretrained(model, **kwargs)
        
        assert tokenizer is not None, "You must give a tokenizer object when you provide a model object (as opposed to just a model name)!"
            
        return model, tokenizer

    def _get_logits(self, token_ids, forced_bytes):
        '''Computes the logits for the given token state.
        
        This overrides a method from the LocalEngine class that is used to get
        inference results from the model.
        '''

        # make sure we don't run off the end of the model
        if len(token_ids) >= getattr(self.model_obj.config, "max_position_embeddings", 1e10):
            raise Exception(f"Attempted to run a transformers model past its maximum context window size of {self.model_obj.config.max_position_embeddings}!")

        # get the number of cache positions we are using
        cache_token_ids = self._cache_state["cache_token_ids"]
        num_cached = 0
        for id in cache_token_ids:
            if num_cached >= len(cache_token_ids) or num_cached >= len(token_ids) or token_ids[num_cached] != id:
                break
            num_cached += 1

        # reset the cache length according to that number of positions
        past_key_values = self._cache_state["past_key_values"]
        past_length = past_key_values[0][0].size(-2) if past_key_values is not None else 0
        if past_length > num_cached:
            past_length = max(0, num_cached - 1) # note we recompute the last token because we don't bother to handle the special case of just computing logits
            self._cache_state["past_key_values"] = tuple(tuple(p[..., :past_length, :] for p in v) for v in past_key_values)
        cache_token_ids[past_length:] = []
        
        # call the model
        new_token_ids = token_ids[past_length:]
        if len(new_token_ids) > 0:
            with torch.no_grad():
                model_out = self.model_obj(
                    input_ids=torch.tensor(new_token_ids).unsqueeze(0).to(self.device),
                    past_key_values=self._cache_state["past_key_values"],
                    use_cache=True,
                    position_ids=torch.arange(past_length, past_length+len(new_token_ids)).unsqueeze(0).to(self.device),
                    attention_mask=torch.ones(1, past_length + len(new_token_ids)).to(self.device),
                    return_dict=True,
                    output_attentions=False,
                    output_hidden_states=False
                )

            # save the results
            self._cache_state["past_key_values"] = model_out.past_key_values
            cache_token_ids.extend(new_token_ids)
            self._cache_state["logits"] = model_out.logits[0, -1, :]
        
        return self._cache_state["logits"]
    
class TransformersChat(Transformers, Chat):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


import pytest
from guidance import select, gen, models, instruction
from ..utils import get_model

def test_palm2_instruct():
    try:
        vmodel = models.VertexAI("text-bison@001")
    except:
        pytest.skip("Skipping OpenAI test because we can't load the model!")

    with instruction():
        lm = vmodel + "this is a test about"
    lm += gen("test", max_tokens=100)
    assert len(lm["test"]) > 0

def test_palm2_chat():
    from guidance import models, gen, system, user, assistant

    try:
        vmodel = models.VertexAI("chat-bison@001")
    except:
        pytest.skip("Skipping OpenAI test because we can't load the model!")

    with system():
        lm = vmodel + "You are an always-happy agent no matter what."

    with user():
        lm += "The economy is crashing!"

    with assistant():
        lm += gen("test1", max_tokens=100)

    with user():
        lm += "What is the best again?"

    with assistant():
        lm += gen("test2", max_tokens=100)

    assert len(lm["test1"]) > 0
    assert len(lm["test2"]) > 0

    # second time to make sure cache reuse is okay
    with system():
        lm = vmodel + "You are an always-happy agent no matter what."

    with user():
        lm += "The economy is crashing!"

    with assistant():
        lm += gen("test1", max_tokens=100)

    with user():
        lm += "What is the best again?"

    with assistant():
        lm += gen("test2", max_tokens=100)

    assert len(lm["test1"]) > 0
    assert len(lm["test2"]) > 0
    assert lm["test1"].find("<|im_end|>") < 0


import os
from pathlib import Path
import multiprocessing
from itertools import takewhile
import operator
import threading
import numpy as np
import queue
import time
import tiktoken
import re

from ._model import Chat, Instruct
from ._local import Local


try:
    # TODO: can we eliminate the torch requirement for llama.cpp by using numpy in the caller instead?
    import torch
    is_torch = True
except ImportError:
    is_torch = False

# try:
#     # TODO: can we eliminate the torch requirement for llama.cpp by using numpy in the caller instead?
#     import vertexai
#     is_vertexai = True
# except ImportError:
#     is_vertexai = False

class Remote(Local):
    def __init__(self, model, tokenizer=None, echo=True, caching=True, temperature=0.0, max_streaming_tokens=500, **kwargs):
        self.caching = caching
        self.temperature = temperature
        self.max_streaming_tokens = max_streaming_tokens

        # Remote models don't always have public tokenizations, so when not provided we pretend they tokenize like gpt2...
        if tokenizer is None:
            tokenizer = tiktoken.get_encoding("gpt2")

        # tiktoken tokenizer was given
        if hasattr(tokenizer, "decode_single_token_bytes"):
            special_map = {v: k for k,v in tokenizer._special_tokens.items()}
            byte_tokens = [b'<|invalid_special_token|>']
            for i in range(tokenizer.n_vocab):
                try:
                    bval = tokenizer.decode_single_token_bytes(i)
                except KeyError:
                    bval = special_map.get(i, b'<|invalid_special_token|>')
                byte_tokens.append(bval)

            bos_token_id = None
            eos_token_id = tokenizer._special_tokens["<|endoftext|>"]
        
        # a transformer tokenizer was given that has a byte_decoder
        elif hasattr(tokenizer, "byte_decoder"):
            byte_tokens = []
            for i in range(tokenizer.vocab_size):
                byte_coded = bytes([tokenizer.byte_decoder[c] for c in tokenizer.convert_ids_to_tokens(i)])
                byte_tokens.append(byte_coded)
            bos_token_id = tokenizer.bos_token_id
            eos_token_id = tokenizer.eos_token_id
        
        # a transformer tokenizer was given with byte_decoder
        elif hasattr(tokenizer, "convert_ids_to_tokens"):
            byte_tokens = [bytes(tokenizer.convert_tokens_to_string(['a', tokenizer.convert_ids_to_tokens(i)])[1:], encoding="utf8") for i in range(tokenizer.vocab_size)]
            bos_token_id = tokenizer.bos_token_id
            eos_token_id = tokenizer.eos_token_id

        else:
            raise Exception("The tokenizer given was not of a recognized type!")

        # build the 
        super().__init__(
            byte_tokens,
            bos_token_id,
            eos_token_id,
            echo=echo
        )

        self._shared_state = {
            "data_queue": queue.Queue(), # this is where the streaming thread puts results
            "data": b'', # these are the bytes we are ready to use in the main thread
            "not_running_stream": threading.Event(), # this is phrased negatively so we can wait for the stop event
            "last_call": 0,
            "num_calls_made": 0
        }
        self._shared_state["not_running_stream"].set()

        self.max_calls = 10
        self.timeout = 50.1

    def _running_stream(self):
        return not self._shared_state["not_running_stream"].is_set() # wrap double negation (which)

    def _start_generator_stream(self, generator):
        dqueue = self._shared_state["data_queue"]
        first_iteration = True
        for chunk in generator:
            if len(chunk) > 0:
                dqueue.put(chunk)
            if self._shared_state["not_running_stream"].is_set() or not first_iteration and time.time() - self._shared_state["last_call"] > self.timeout:
                break
            first_iteration = False

        if self._running_stream():
            dqueue.put(self.tokens[self.eos_token_id])
        self._shared_state["not_running_stream"].set()
        dqueue.put(b'') # so we never get stuck waiting for a running stream to return something

    def _start_new_stream(self, prompt):

        if self._shared_state["num_calls_made"] > self.max_calls:
            raise Exception(f"We have exceeded the maximum number of calls! {self.max_calls}")

        # stop any running stream
        if self._running_stream():
            self._shared_state["not_running_stream"].set() # stop the generator
            self._shared_state["remote_thread"].join() # wait for the thread to finish

        # clear the data queue
        while not self._shared_state["data_queue"].empty(): 
            self._shared_state["data_queue"].get()

        # start the new stream
        self._shared_state["last_call"] = time.time()
        generator = self._generator(prompt)
        self._shared_state["not_running_stream"].clear() # so we know we are running
        self._shared_state["num_calls_made"] += 1

        self._shared_state["remote_thread"] = threading.Thread(target=self._start_generator_stream, args=(generator,))
        self._shared_state["remote_thread"].start()
        # self._start_generator_stream(stripped_prompt)
    
    def _get_logits(self, token_ids, forced_bytes):
        '''Computes the logits for the given token state.
        
        This overrides a method from the Local class that is used to get
        inference results from the model.
        '''

        if len(token_ids) == 0:
            raise ValueError("token_ids must contain some tokens.")
        
        # compute the prompt bytes
        prompt = b''.join([self.tokens[i] for i in token_ids]) + forced_bytes

        self._shared_state["last_call"] = time.time()

        # keep looping until we have at least one more byte past our prompt
        token_id = None
        while True:

            # try and get the next token id
            if self._shared_state["data"].startswith(prompt):
                token_id = self._get_next_token(len(prompt)-len(forced_bytes))
                if token_id is not None:
                    break

            # restart if extending our data will never lead to matching our prompt
            elif len(self._shared_state["data"]) >= len(prompt) or not prompt.startswith(self._shared_state["data"]):
                self._start_new_stream(prompt)

            # extend our data with a chunk from the model stream
            if not self._shared_state["data_queue"].empty():
                new_bytes = self._shared_state["data_queue"].get_nowait()
                
                # if we are at the end of the generation then we try again allowing for early token stopping
                if len(new_bytes) == 0:
                    token_id = self._get_next_token(len(prompt), allow_early_stop=True)
                    if token_id is not None:
                        break
                self._shared_state["data"] += new_bytes
            
            # but if there is nothing and we are not running then we start a stream
            elif self._shared_state["not_running_stream"].is_set():
                self._start_new_stream(prompt)

            # we wait for the running stream to put something in the queue
            else:
                self._shared_state["data"] += self._shared_state["data_queue"].get()
        
        # # if we don't have the next byte of data yet then we wait for it (from the streaming thread)
        # if len(self._shared_state["data"]) == len(prompt):
        #     self._shared_state["data"] += self._shared_state["data"]_queue.get() 

        # token_id = self._get_next_token(len(prompt))

        # set the logits to the next byte the model picked
        logits = np.ones(len(self.tokens)) * np.nan
        logits[token_id] = 100
        
        return torch.tensor(logits) # TODO: the caller need to know to NOT use the 0 entries, but to fail out if that happens
    
    def _get_next_token(self, pos, allow_early_stop=False):
        data = self._shared_state["data"]
        trie = self._token_trie
        token_id = None
        while True:
            
            # see if we have run out of data
            if pos >= len(data):
                if allow_early_stop:
                    return token_id
                else:
                    return None
            
            # try and walk down the trie
            next_byte = data[pos:pos+1]
            if next_byte in trie.children:
                trie = trie.children[next_byte]
                pos += 1
                if trie.value is not None:
                    token_id = trie.value
            else:
                return token_id # this is the longest greedy token match we can make


import guidance
from guidance import gen, models, commit_point, Tool, select, capture, string
from ..utils import get_model
import re

def test_basic():
    lm = models.LocalMock()
    lm += "Write a number: " + gen('text', max_tokens=3)
    assert len(lm["text"]) > 0

def test_stop_string():
    lm = models.LocalMock(b"<s>Count to 10: 1, 2, 3, 4, 5, 6, 7, 8, 9, 10")
    lm += "Count to 10: 1, 2, 3, 4, 5, 6, 7, " + gen('text', stop=", 9")
    assert lm["text"] == "8"

def test_stop_char():
    lm = models.LocalMock(b"<s>Count to 10: 1, 2, 3, 4, 5, 6, 7, 8, 9, 10")
    lm += "Count to 10: 1, 2, 3, 4, 5, 6, 7, " + gen('text', stop=",")
    assert lm["text"] == "8"

def test_stop_quote():
    lm = get_model("transformers:gpt2")
    lm += '''A title: "''' + gen('title', max_tokens=30, stop='"')
    assert not lm["title"].endswith('"')

def test_unicode():
    lm = get_model("transformers:gpt2")
    lm + '''Question: Josh decides to try flipping a house.  He buys a house for $80,000 and then puts in $50,000 in repairs.  This increased the value of the house by 150%.  How much profit did he make?
Let's think step by step, and then write the answer:
Step 1''' + gen('steps', list_append=True, stop=['\nStep', '\n\n', '\nAnswer'], temperature=0.7, max_tokens=20) + '\n'
    i = 2
    lm + f'Step {i}:' + gen('steps', list_append=True, stop=['\nStep', '\n\n', '\nAnswer'], temperature=0.7, max_tokens=20) + '\n'

def test_unicode2():
    lm = get_model("transformers:gpt2")
    prompt = 'Janet’s ducks lay 16 eggs per day'
    lm +=  prompt + gen(max_tokens=10)
    assert True

def test_gsm8k():
    lm = models.LocalMock()
    lm + '''Question: Janet's ducks lay 16 eggs per day. She eats three for breakfast every morning and bakes muffins for her friends every day with four. She sells the remainder at the farmers' market daily for $2 per fresh duck egg. How much in dollars does she make every day at the farmers' market?
Answer: ''' + gen(max_tokens=30)
    assert True

def test_pattern_kleene():
    lm = get_model("transformers:gpt2")
    lm += 'The Lord is my'
    x = lm + gen(name='tmp', max_tokens=10)
    y = lm + gen(name='tmp', regex='.*', max_tokens=10)
    assert x['tmp'] == y['tmp']

def test_non_token_force():
    '''This forces some bytes that don't match a token (only longer tokens)'''
    lm = get_model("transformers:gpt2")
    lm += 'ae ae' + gen(regex='\d')
    assert len(str(lm)) == 6

def test_gen_vs_grammar():
    lm = get_model("transformers:gpt2")
    lm += 'The Lord is my'
    x = lm + gen(name='tmp', max_tokens=10)
    y = lm + gen(name='tmp',  regex='.*', max_tokens=10)
    assert x['tmp'] == y['tmp']

def test_pattern_optional():
    lm = models.LocalMock(b"<s>12342333")
    pattern = '.?233'
    lm2 = lm + '123' + gen(name='numbers', regex=pattern, max_tokens=10)
    assert lm2['numbers'] == '4233'
    lm = models.LocalMock(b"<s>1232333")
    pattern = '.?233'
    lm2 = lm + '123' + gen(name='numbers', regex=pattern, max_tokens=10)
    assert lm2['numbers'] == '233'
    pattern = '(Scott is bad)?(\d+)?o'
    lm = models.LocalMock(b"<s>John was a little man full of things")
    lm2 = lm + 'J' + gen(name='test', regex=pattern, max_tokens=30)
    assert lm2['test'] == 'o'

def test_pattern_stops_when_fulfilled():
    lm = models.LocalMock(b"<s>123abc")
    lm += gen(regex='\d+', max_tokens=10, name='test')
    assert lm['test'] == '123'

def test_pattern_star():
    # lm = models.LocalMock(b"<s>1234233234<s>") # commented out because it is not a valid test
    # patterns = ['\d+233', '\d*233', '.+233', '.*233']
    # for pattern in patterns:
    #     lm2 = lm + '123' + gen(name='numbers', regex=pattern, max_tokens=10)
    #     assert lm2['numbers'] == '4233'
    lm = models.LocalMock(b"<s>123233")
    patterns = ['\d*233','.*233']
    for pattern in patterns:
        lm2 = lm + '123' + gen(name='numbers', regex=pattern, max_tokens=10)
        assert lm2['numbers'].startswith('233')
    pattern = '.*(\n|little)'
    lm = models.LocalMock(b"<s>John was a little")
    lm2 = lm + 'J' + gen(name='test', regex=pattern, max_tokens=30)
    assert lm2['test'].startswith('ohn was a little')
    lm = models.LocalMock(b"<s>John was a litt\n")
    lm2 = lm + 'J' + gen(name='test', regex=pattern, max_tokens=30)
    assert lm2['test'].startswith('ohn was a litt\n')

def test_stop_regex():
    lm = models.LocalMock(b"<s>123a3233")
    lm2 = lm + '123' + gen(name='test', stop_regex='\d233', max_tokens=10)
    assert lm2['test'] == 'a'
    lm = models.LocalMock(b"<s>123aegalera3233")
    lm2 = lm + '123' + gen(name='test', stop_regex='\d', max_tokens=30)
    assert lm2['test'] == 'aegalera'

def test_stop_regex_star():
    lm = models.LocalMock(b"<s>123a3233")
    pattern = '\d+233'
    lm2 = lm + '123' + gen(name='test', stop_regex=pattern, max_tokens=10)
    assert lm2['test'] == 'a'

def test_empty_pattern():
    pattern = '(Scott is bad)?(\d+)?'
    lm = models.LocalMock(b"<s>J<s>")
    lm2 = lm + 'J' + gen(name='test', regex=pattern, max_tokens=30)
    assert lm2['test'] == ''

def test_various_regexes():
    lm = get_model("transformers:gpt2")
    prompts = ['Hi there', '2 + 2 = ', 'Scott is a', 'I have never seen a more', 'What is the', '?FD32']
    patterns = ['(Scott is a person|Scott is a persimmon)', 'Scott is a persimmon.*\.', '\d\.*\d+']
    for prompt in prompts:
        for pattern in patterns:
            lm2 = lm + prompt + gen(name='test', regex=pattern, max_tokens=40)
            assert re.match(pattern, lm2['test']) is not None # note we can't just test any regex pattern like this, we need them to have finished in less than 40 tokens
# def test_pattern():
#     lm = get_model("transformers:gpt2")
#     lm += 'hey there my friend what is truth 23+43=' + gen(regex=r'dog(?P<stop>.+)', max_tokens=30)
#     assert str(lm) == "hey there my friend what is truth 23+43=dog"

def test_long_prompt():
    lm = get_model("transformers:gpt2")
    prompt = '''Question: Legoland has 5 kangaroos for each koala. If Legoland has 180 kangaroos, how many koalas and kangaroos are there altogether?
Let's think step by step, and then write the answer:
Step 1: For every 5 kangaroos, there is one koala, meaning for the 180 kangaroos, there are 180/5 = 36 koalas.
Step 2: Altogether, there are 36+180 = 216 koalas and kangaroos.
Answer: 216

Question: Jennifer has ten pears, 20 oranges, and twice as many apples as pears. If she gives her sister two of each fruit, how many fruits does she have left?
Let's think step by step, and then write the answer:
Step 1: Jennifer has 2*10 = 20 apples.
Step 2: She has a total of 10+20+20 = 50 fruits.
Step 3: She gives her sister 2 pears+2 oranges+2 apples = 2+2+2 = 6 fruits.
Step 4: After giving her sister 2 of each fruit, she has 50-6 = 44 fruits left.

Answer: 44

Question: A busy restaurant is counting how many customers they had during that Friday to try to predict how many they might get on Saturday. During breakfast, they had 73 customers. During lunch, they had 127 customers. During dinner, they had 87 customers. If they predict they'll get twice the amount of customers on Saturday as they had on Friday, how many customers do they predict they will get?
Let's think step by step, and then write the answer:
Step 1: On Friday the restaurant had 73 customers for breakfast + 127 customers for lunch + 87 customers for dinner = 73+127+87 = 287 customers total on Friday.
Step 2: If they predict getting 2x the amount of customers on Saturday as on Friday, they will have 287 customers x 2 = 287*2 = 574 customers on Saturday.

Answer: 574

Question: They say the first year of a dog's life equals 15 human years. The second year of a dog's life equals 9 human years and after that, every year of a dog's life equals 5 human years. According to this logic, how many human years has my 10-year-old dog lived?
Let's think step by step, and then write the answer:
Step 1: If your dog is 10 years old then in his first year of life he lived 1*15 = 15 human years
Step 2: In his second year of life, he lived 1*9 = 9 human years
Step 3: We need to calculate his remaining years or 10-2 = 8 years of dog life into human years
Step 4: If 1 year of dog life after the 2 years equates to 5 human years, then 8 years of dog life equals 8*5 = 40 human years
Step 5: In total, your dog has lived 15 + 9 + 40 = 64 human years

Answer: 64

Question: Amanda and her family are going to re-paint all the walls inside their house. Before they get started they want to divide up the work. Since all the rooms in the house have different numbers and sizes of walls in them, they figure the fairest way to divide up the work is to count all the walls in the house and assign an equal number to each person. There are 5 people in Amanda's family, including herself. There are 9 rooms in the house. 5 of the rooms have 4 walls each. The other 4 rooms each have 5 walls each. To be fair, how many walls should each person in Amanda's family paint?
Let's think step by step, and then write the answer:
Step 1: First, Amanda needs to figure out how many walls there are in the house, 5 rooms x 4 walls each = 5*4 = 20.
Step 2: The other 4 rooms have x 5 walls each = 4 * 5 = 20 walls.
Step 3: The house has 20 walls + 20 walls = 20 + 20 = 40 walls total.
Step 4: To divide the work equally between the 5 people, 40 walls / 5 people = 40/5 = 8 walls for each person.

Answer: 8

Question: Janet’s ducks lay 16 eggs per day. She eats three for breakfast every morning and bakes muffins for her friends every day with four. She sells the remainder at the farmers' market daily for $2 per fresh duck egg. How much in dollars does she make every day at the farmers' market?
Let's think step by step, and then write the answer:
Step 1'''
    lm +=  prompt + gen(max_tokens=10)
    assert True

def test_list_append():
    '''This tests is list append works across grammar appends.'''
    lm = models.LocalMock(b"<s>bababababa")
    lm += "<s>"
    for _ in range(3):
        lm += gen("my_list", list_append=True, stop="a") + "a"
    assert isinstance(lm['my_list'], list)
    assert len(lm['my_list']) == 3

def test_list_append_in_grammar():
    '''This tests is list append works within the same grammar.'''
    lm = models.LocalMock(b"<s>bababababa")
    lm += "<s>"
    lm += gen("my_list", list_append=True, stop="a") + "a" + gen("my_list", list_append=True, stop="a") + "a" + gen("my_list", list_append=True, stop="a")
    assert isinstance(lm['my_list'], list)
    assert len(lm['my_list']) == 3

def test_one_char_suffix_and_regex():
    model = models.LocalMock(b"<s>this is\na test")
    model += gen(regex=".*", suffix="\n", max_tokens=20)
    assert str(model) == "this is\n"

def test_one_char_stop_and_regex():
    model = models.LocalMock(b"<s>this is\na test")
    model += gen(regex=".*", stop="\n", max_tokens=20)
    assert str(model) == "this is"


from guidance import models, block, any_char

def test_text_opener():
    model = models.LocalMock("<s>open texta")
    with block(opener="open text"):
        model += any_char()
    assert str(model) == "open texta"

def test_text_closer():
    model = models.LocalMock("<s>aclose text")
    model += "<s>"
    with block(closer="close text"):
        model += any_char()
    assert str(model) == "<s>aclose text"

def test_grammar_opener():
    model = models.LocalMock("<s>open texta")
    with block(opener="open tex" + any_char()):
        model += any_char()
    assert str(model) == "open texta"

def test_grammar_closer():
    model = models.LocalMock(["<s>aclose text", "<s>close text"])
    model += "<s>"
    try:
        with block(closer=any_char() + "lose text"):
            model += any_char()
    except:
        return # we expect an exception
    assert False, "We should have thrown an exception using a context (prompt) based grammar in the closer!"


from guidance import commit_point, models, Tool, select, capture, string
from ..utils import get_model

def test_hidden():
    model = models.LocalMock()
    model += " one" + commit_point(" two", hidden=True) + " three"
    assert str(model) == " one three"

def test_commit_point():
    lm = get_model("transformers:gpt2")
    tools = [Tool(callable=lambda x: x)]
    stop_pattern = select([string(lm.eos_token)])
    gen_grammar = select([commit_point(stop_pattern, hidden=True)] + [capture(commit_point(x.call_grammar, hidden=True), name=f'tool{i}') for i, x in enumerate(tools)])


import numpy as np
from ordered_set import OrderedSet
from ._grammar import Join, Select, Terminal, Null, Byte, ByteRange

class EarleyItem:
    __slots__ = ("node", "values", "start", "pos", "log_prob", "children", "hidden_start")

    def __init__(self, node, values, pos, start, log_prob, hidden_start):
        self.node = node
        self.values = values
        self.start = start
        self.pos = pos
        self.log_prob = log_prob
        self.children = None
        self.hidden_start = hidden_start

    def __eq__(self, other):
        return isinstance(other, EarleyItem) and \
               self.start == other.start and \
               self.pos == other.pos and \
               self.node == other.node and \
               self.values == other.values and \
               self.log_prob == other.log_prob
    
    def __hash__(self):
        return hash((self.node, self.values, self.start, self.pos))
    
    def __repr__(self):
        if isinstance(self.node, Join):
            s = f"{self.node.name:20} -> "
            rs = ""
            for i,v in enumerate(self.values):
                if self.pos == i:
                    rs += "•"
                rs += v.name + " "
            if self.pos == len(self.values):
                rs += "•"
        elif isinstance(self.node, Select):
            s = f"{self.node.name:20} -> "
            rs = ""
            if self.pos == 0:
                rs += "•"
            rs += self.values[0].name
            if self.pos == 1:
                rs += "•"
        else:
            assert False
        return s + f"{rs:40} ({self.start}) {'nullable' if self.node.nullable else ''}"

class EarleyCommitParser:
    def __init__(self, grammar):

        # we can't have a terminal as the root
        if isinstance(grammar, Terminal):
            grammar = Join([grammar])
        
        self.grammar = grammar
        self.bytes = b''
        self.state_sets = [OrderedSet()] # the list of Earley items for each byte
        self.token_counts = [] # used to track how many tokens have been used
        self.state_set_pos = 0
        self.shadow_pos = 0
        self._add_node(self.grammar, 0, 0.0, 1000000000)
        self._inner_loop(self.state_set_pos)

    @property
    def pos(self):
        return self.shadow_pos
    @pos.setter
    def pos(self, new_pos):

        # do nothing if we aren't moving
        if new_pos == self.state_set_pos:
            return
        elif new_pos > self.state_set_pos:
            raise Exception("Can't move the parser position forward! (only backward)")
        
        # check if we are just moving the shadow position
        if new_pos >= self.shadow_pos:
            self.shadow_pos = new_pos
            return
        
        # actually reset our position if we need to
        self.state_sets = self.state_sets[:new_pos+1] + [OrderedSet()]
        self.token_counts = self.token_counts[:new_pos+2]
        self.bytes = self.bytes[:new_pos]
        self.state_set_pos = new_pos
        self.shadow_pos = new_pos
        self._inner_loop(self.state_set_pos)

    def _add_item(self, state_set_pos, new_item):
        state_set = self.state_sets[state_set_pos]
        if new_item not in state_set:
            state_set.append(new_item)
        else:
            existing_item = state_set.items[state_set.map[new_item]]
            existing_item.hidden_start = min(existing_item.hidden_start, new_item.hidden_start)

    def _add_node(self, grammar, state_set_pos, log_prob, hidden_start):
        if isinstance(grammar, Terminal):
            new_item = EarleyItem(grammar, tuple(), 0, state_set_pos, log_prob, hidden_start)
            self._add_item(state_set_pos, new_item)
            
        elif isinstance(grammar, Join):
            new_item = EarleyItem(grammar, tuple(grammar.values), 0, state_set_pos, log_prob, hidden_start)
            self._add_item(state_set_pos, new_item)
        
        elif isinstance(grammar, Select):
            for value in grammar.values:
                new_item = EarleyItem(grammar, (value,), 0, state_set_pos, log_prob, hidden_start)
                self._add_item(state_set_pos, new_item)

    def _inner_loop(self, state_set_pos, start_pos=0):
        curr_state_set = self.state_sets[state_set_pos]
        if len(self.state_sets) == state_set_pos + 1:
            self.state_sets.append(OrderedSet())
            self.token_counts.append(self.token_counts[-1] if len(self.token_counts) > 0 else 0)
        next_state_set = self.state_sets[state_set_pos + 1]
        pos = start_pos
        while len(curr_state_set) > pos:
            item = curr_state_set[pos]

            # completion
            if item.pos == len(item.values):
                
                # if we complete an item that is a "commit point" then we eliminate all other possible
                # parses so that we are "committed" to using this item
                # we do this by removing any unprocessed items in the current state set and clearing the next state set
                if item.node.commit_point:
                    while len(curr_state_set) > pos + 1:
                        curr_state_set.pop()
                    next_state_set.clear()
                
                # advance all the parents that our completion impacts
                token_span = self.token_counts[state_set_pos] - self.token_counts[item.start]
                start_state_set = self.state_sets[item.start]
                for start_item in start_state_set:
                    if start_item.pos < len(start_item.values) and start_item.values[start_item.pos] == item.node:
                        
                        if item.node.max_tokens <= token_span and any(start_item.node == v  for v in item.node.values):
                            continue # skip advancing parents that are also children (recursion) once we are past the token limit

                        curr_state_set.append(EarleyItem(
                            start_item.node,
                            start_item.values,
                            start_item.pos + 1,
                            start_item.start,
                            start_item.log_prob + item.log_prob, # increment the log prob by the child value,
                            start_item.hidden_start
                        ))
            
            # don't advance past our max token limit
            elif item.node.max_tokens > self.token_counts[state_set_pos] - self.token_counts[item.start]:

                # scan (note we only scan forward when we have more max token headroom left)
                next_item_node = item.values[item.pos]
                hidden_start = item.hidden_start
                if next_item_node.hidden:
                    hidden_start = min(state_set_pos, hidden_start)
                if isinstance(next_item_node, Terminal):# and item.node.max_tokens > self.token_counts[state_set_pos] - self.token_counts[item.start]:
                    next_state_set.append(EarleyItem(item.node, item.values, item.pos + 1, item.start, item.log_prob, hidden_start)) # the log prob will get incremented when consume_bytes is called
                
                # prediction
                else:
                    self._add_node(next_item_node, state_set_pos, 0.0, hidden_start) # the log probs will get incremented by children later

                # handle nullable items by advancing them automatically (since we know we can)
                if next_item_node.nullable:
                    new_item = EarleyItem(item.node, item.values, item.pos + 1, item.start, item.log_prob, item.hidden_start)
                    if new_item not in self.state_sets[state_set_pos]:
                        self.state_sets[state_set_pos].append(new_item)
            pos += 1

    def earliest_hidden_start(self, state_pos=None):
        '''The earliest that a hidden node might match.
        
        This is useful because it tells us which bytes may end being hidden.
        '''
        if state_pos is None:
            state_pos = self.state_set_pos
        earliest_pos = 10000000000
        for item in self.state_sets[state_pos]:
            earliest_pos = min(earliest_pos, item.hidden_start)
        return earliest_pos
    
    # def earliest_hidden_start(self, state_pos=None):
    #     '''The earliest that a hidden node might match.
        
    #     This is useful because it tells us which bytes may end being hidden.
    #     '''
    #     if state_pos is None:
    #         state_pos = self.state_set_pos
    #     earliest_pos = 10000000000
    #     for item in self.state_sets[state_pos]:

    #         if item.pos > 0:
    #             # check for hidden nodes
    #             if item.node.hidden and item.start < earliest_pos:
    #                 earliest_pos = item.start
                
    #             # check for nodes that are not hidden but end with a hidden terminal (terminal won't be in our state set by themselves, so we need this check)
    #             else:
    #                 last_value = item.values[item.pos-1]
    #                 if isinstance(last_value, Terminal) and last_value.hidden and state_pos - len(last_value) < earliest_pos:
    #                     earliest_pos = state_pos - len(last_value)
        
    #     return earliest_pos
    
    def matched(self):
        '''Checks if the parser has completely matched the grammar.'''
        if self.shadow_pos != self.state_set_pos:
            return False
        for item in self.state_sets[self.state_set_pos]:
            if item.node == self.grammar and item.pos == len(item.values):
                return True
        return False
    
    def shadow_rewind(self, new_pos):
        if new_pos == self.state_set_pos:
            return
        self.shadow_pos = new_pos
    
    def commit_and_collapse_item(self, item):
        '''This collapses the item into zero width and rewinds the parser position accordingly.
        
        Note we assume the item is in the current state set.
        '''

        # trim off the state sets that matches this item
        self.state_sets = self.state_sets[:item.start + 1]
        self.token_counts = self.token_counts[:item.start + 1]
        self.bytes = self.bytes[:item.start]
        self.state_set_pos = item.start
        self.shadow_pos = item.start

        # add this state to its start point (making it a zero length match with no values)
        self.state_sets[item.start].append(EarleyItem(item.node, tuple(), 0, item.start, item.log_prob, item.hidden_start))

        # expand from this state
        self._inner_loop(item.start, len(self.state_sets[item.start]) - 1)

    def mark_new_token(self):
        # TODO: we allow ourselves to go one past our max token limit when we hit a one-byte token
        #       because we don't know if we are continuing or extending a new token when we parse
        #       the first byte of the token. We could fix this by rerunning the inner_loop after each
        #       token, but we skip that for now since max_tokens is not a hard garuntee anyway when you
        #       have patterns.
        
        self.token_counts[-1] += 1

    def consume_byte(self, byte, log_prob=0.0):
        '''Advances the parser by the given byte.'''

        # see if we need to advance our shadow position...
        if self.shadow_pos < self.state_set_pos:
            assert byte == self.bytes[self.shadow_pos:self.shadow_pos+1], "Attempted to consume a byte by advancing shadow_pos but the byte didn't match!"
            self.shadow_pos += 1
            return

        # ...if not, we extend our bytes
        self.bytes += byte

        # filter out all the extensions that don't match this byte
        new_next_state_set = []
        found_valid = False
        found_invalid = False
        hidden_start = 10000000000
        for item in self.state_sets[self.state_set_pos + 1]:
            if item.pos > 0 and isinstance(item.values[item.pos - 1], Terminal):
                last_inner_node = item.values[item.pos - 1]
                if not last_inner_node.match_byte(byte):
                    found_invalid = True
                    continue
                else:
                    found_valid = True
                    if last_inner_node.commit_point:
                        item.log_prob += log_prob
                        new_next_state_set = [item]
                        hidden_start = min(hidden_start, item.hidden_start)
                        found_invalid = True # we make everything else invalid, so that means we found something invalid
                        break
            item.log_prob += log_prob # update the probability of the item by the probability of choosing this byte
            new_next_state_set.append(item)
            hidden_start = min(hidden_start, item.hidden_start)
        if not found_valid:
            raise Exception("Attempted to consume a byte that the grammar does not accept!")
        if found_invalid: # only update if we changed the set
            self.state_sets[self.state_set_pos + 1] = OrderedSet(new_next_state_set)

        # advance the parser one position
        self.state_set_pos += 1
        self.shadow_pos += 1
        self._inner_loop(self.state_set_pos)

        # look for a commit point node
        commit_point = None
        for item in self.state_sets[self.state_set_pos]:
            if item.node.commit_point and item.pos == len(item.values) or (item.pos > 0 and item.values[item.pos-1].commit_point):
                commit_point = item
                break # TODO: consider how we might need to prioritize multiple commit point nodes (an uncommon scenario I think)
        # hidden_start, 
        return commit_point

    def valid_next_bytes(self):
        '''A list of Byte and ByteRange objects representing the next valid bytes.'''
        valid_items = set()
        next_state_set = self.state_sets[self.state_set_pos + 1]
        for item in next_state_set:
            if item.pos > 0 and isinstance(item.values[item.pos - 1], Terminal):
                v = item.values[item.pos - 1]
                if v not in valid_items:
                    valid_items.add(v)
        return valid_items
    
    def next_byte_temperature(self):
        '''The maximum temperature over all the next bytes, or -1 if no temperature is set.'''
        max_temp = -1
        next_state_set = self.state_sets[self.state_set_pos + 1]
        for item in next_state_set:
            if item.pos > 0 and isinstance(item.values[item.pos - 1], Terminal):
                v = item.values[item.pos - 1]
                max_temp = max(max_temp, v.temperature)
        return max_temp
    
    def next_byte_mask(self):
        '''A mask version of the `valid_next_bytes` method.'''
        
        mask = np.zeros(256, dtype=bool)

        # if we are shadow rewound then we just force those bytes again
        if self.shadow_pos < self.state_set_pos:
            mask[self.bytes[self.shadow_pos]] = True
        
        # otherwise we compute the valid bytes from the grammar
        else:
            valid_items = self.valid_next_bytes()
            for item in valid_items:
                if isinstance(item, Byte):
                    mask[item.byte[0]] = True
                elif isinstance(item, ByteRange):
                    mask[item.byte_range[0]:item.byte_range[1]+1] = True
                else:
                    raise Exception("Unknown Terminal Type: "  + str(type(item)))
        return mask

    def __repr__(self, state_sets=None) -> str:
        s = ""
        if state_sets is None:
            state_sets = self.state_sets
        for i,states in enumerate(state_sets):
            s += f"\n=== {i} ==="
            if self.state_set_pos == i:
                s += " (state_set_pos)"
            s += "\n"
            for state in states:
                if isinstance(state.node, Join):
                    s += f"{state.node.name:20} -> "
                    rs = ""
                    for i,v in enumerate(state.values):
                        if state.pos == i:
                            rs += "•"
                        rs += v.name + " "
                    if state.pos == len(state.values):
                        rs += "•"
                elif isinstance(state.node, Select):
                    s += f"{state.node.name:20} -> "
                    rs = ""
                    if state.pos == 0:
                       rs += "•"
                    rs += state.values[0].name
                    if state.pos == 1:
                       rs += "•"
                else:
                    assert False
                s += f"{rs:40} ({state.start}) {'nullable' if state.node.nullable else ''}\n"
        return s
    
    def _reversed_state_sets(self):
        new_state_sets = [OrderedSet([]) for _ in range(len(self.state_sets))]
        for i,states in enumerate(self.state_sets):
            for state in states:
                # if state.node.name == "__call___c":
                #     pass
                new_state_sets[state.start].append(EarleyItem(state.node, state.values, state.pos, i, state.log_prob, state.hidden_start))
        
        return new_state_sets
    
    def parse_tree(self):
        reversed_state_sets = self._reversed_state_sets()

        # find the matching root state
        for item in reversed_state_sets[0]:
            if item.node == self.grammar and item.start == len(self.bytes) and item.pos == len(item.values): # note that ".start" mean end because items are reversed
                root_item = item
        self._compute_parse_tree(0, root_item, reversed_state_sets)
        return root_item
    
    def _compute_parse_tree(self, initial_pos, initial_item, reversed_state_sets):
        stack = [(initial_pos, initial_item)]
        
        while stack:
            pos, item = stack.pop()

            # compute the children for this item
            assert self._compute_children(pos, item, reversed_state_sets)

            # recurse on the children
            for child in item.children:
                if child is None:
                    pass # this child was nullable and was chosen to be null (empty)
                elif isinstance(child, Terminal):
                    pos += len(child)
                else:
                    stack.append((pos, child))
                    pos = child.start # note that ".start" mean end because items are reversed

    def _compute_children(self, state_set_pos, item, reversed_state_sets, values_pos = 0):

        # ensure we have a children array
        if item.children is None:
            item.children = [None for _ in range(len(item.values))]

        # consume as many terminal children as possible
        while True:
            
            # if we are at the end of the values then there no more children and we see if we consumed all the right bytes
            if values_pos == len(item.values):
                return state_set_pos == item.start # note that ".start" mean end because items are reversed

            # get the child we are trying to match (meaning we are looking for completed early items for this node)
            value = item.values[values_pos]

            # if we have a terminal node we can jump forward that many bytes
            if isinstance(value, Terminal):
                item.children[values_pos] = value
                values_pos += 1
                state_set_pos += len(value)
            else:
                break
            
        # otherwise we need to try all possible next matching items in the current state set
        # so we loop over every item in the current state set looking for a completed match
        for inner_item in reversed_state_sets[state_set_pos]:
            if inner_item.node == value and inner_item.pos == len(inner_item.values):

                # see if we can get a complete parse following this inner item
                if self._compute_children(inner_item.start, item, reversed_state_sets, values_pos + 1):
                    item.children[values_pos] = inner_item
                    return True
                    
        # if we didn't find a child set and this is nullable we can skip this child (since it may not exist if nulled)
        if value.nullable:
            if self._compute_children(state_set_pos, item, reversed_state_sets, values_pos + 1):
                item.children[values_pos] = None # this child was skipped since it was nullable
                return True
        
        return False




import os
from pathlib import Path
import multiprocessing
from itertools import takewhile
import operator
import threading
import numpy as np
import queue
import time
import tiktoken
import re

from ._model import Chat, Instruct
from ._remote import Remote


try:
    # TODO: can we eliminate the torch requirement for llama.cpp by using numpy in the caller instead?
    import torch
    is_torch = True
except ImportError:
    is_torch = False

try:
    # TODO: can we eliminate the torch requirement for llama.cpp by using numpy in the caller instead?
    import openai as openai_package
    is_openai = True
except ImportError:
    is_openai = False

chat_model_pattern = r'^(gpt-3\.5-turbo|gpt-4)(-\d+k)?(-\d{4})?$'

class OpenAI(Remote):
    def __init__(self, model, tokenizer=None, echo=True, caching=True, api_key=None, organization=None, base_url=r"https://api.openai.com/v1", temperature=0.0, max_streaming_tokens=1000, **kwargs):
        if not is_openai or not hasattr(openai_package, "OpenAI"):
            raise Exception("Please install the openai package version >= 1 using `pip install openai -U` in order to use guidance.models.OpenAI!")
        
        # if we are called directly (as opposed to through super()) then we convert ourselves to a more specific subclass if possible
        if self.__class__ is OpenAI:
            found_subclass = None
            # from . import openai

            # if isinstance(model, str):
            #     model_name = model
            # else:
            #     model_name = self.model_obj._model_id

            # chat
            if re.match(chat_model_pattern, model):
                found_subclass = OpenAIChat

            # instruct
            elif "instruct" in model:
                found_subclass = OpenAIInstruct

            # regular completion
            else:
                found_subclass = OpenAICompletion
            
            # convert to any found subclass
            self.__class__ = found_subclass
            found_subclass.__init__(self, model, tokenizer=tokenizer, echo=echo, caching=caching, api_key=api_key, organization=organization, base_url=base_url, temperature=temperature, max_streaming_tokens=max_streaming_tokens, **kwargs)
            return # we return since we just ran init above and don't need to run again

        # Configure an AsyncOpenAI Client with user params.
        if api_key is None:
            api_key = os.environ.get("OPENAI_API_KEY")

        if organization is None:
            organization = os.environ.get("OPENAI_ORG_ID")

        self.client = openai_package.OpenAI(api_key=api_key, organization=organization, base_url=base_url)
        self.model_name = model

        
        
        # self.tokenizer = tiktoken.encoding_for_model(model)
        # self.eos_token = b"<|endoftext|>"

        super().__init__(
            model, tokenizer=tiktoken.encoding_for_model(model), echo=echo,
            caching=caching, temperature=temperature,
            max_streaming_tokens=max_streaming_tokens, **kwargs
        )
        
    

class OpenAICompletion(OpenAI, Instruct):

    def _generator(self, prompt):
        self._shared_state["not_running_stream"].clear() # so we know we are running
        self._shared_state["data"] = prompt # we start with this data

        try:
            generator = self.client.completions.create(
                model=self.model_name,
                prompt=prompt.decode("utf8"), 
                max_tokens=self.max_streaming_tokens, 
                n=1, 
                top_p=1, 
                temperature=0, 
                stream=True
            )
        except Exception as e: # TODO: add retry logic
            raise e
        
        for part in generator:
            chunk = part.choices[0].text or ""
            yield chunk.encode("utf8")

class OpenAIInstruct(OpenAI, Instruct):

    def get_role_start(self, name):
        return ""
    
    def get_role_end(self, name):
        if name == "instruction":
            return "<|endofprompt|>"
        else:
            raise Exception(f"The OpenAIInstruct model does not know about the {name} role type!")

    def _generator(self, prompt):
        # start the new stream
        prompt_end = prompt.find(b'<|endofprompt|>')
        if prompt_end >= 0:
            stripped_prompt = prompt[:prompt_end]
        else:
            raise Exception("This model cannot handle prompts that don't match the instruct format!")
        
        # make sure you don't try and instruct the same model twice
        if b'<|endofprompt|>' in prompt[prompt_end + len(b'<|endofprompt|>'):]:
            raise Exception("This model has been given two separate instruct blocks, but this is not allowed!")
        
        self._shared_state["not_running_stream"].clear() # so we know we are running
        self._shared_state["data"] = stripped_prompt + b'<|endofprompt|>'# we start with this data

        try:
            generator = self.client.completions.create(
                model=self.model_name,
                prompt=self._shared_state["data"].decode("utf8"), 
                max_tokens=self.max_streaming_tokens, 
                n=1, 
                top_p=1, 
                temperature=0, 
                stream=True
            )
        except Exception as e: # TODO: add retry logic
            raise e
        
        for part in generator:
            chunk = part.choices[0].text or ""
            yield chunk.encode("utf8")

class OpenAIChat(OpenAI, Chat):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _generator(self, prompt):
        
        # find the system text
        pos = 0
        # system_start = b'<|im_start|>system\n'
        # user_start = b'<|im_start|>user\n'
        # assistant_start = b'<|im_start|>assistant\n'
        role_end = b'<|im_end|>'
        # system_start_pos = prompt.startswith(system_start)
        
        # find the system text
        # system_text = b''
        # if prompt.startswith(system_start):
        #     pos += len(system_start)
        #     system_end_pos = prompt.find(role_end)
        #     system_text = prompt[pos:system_end_pos]
        #     pos = system_end_pos + len(role_end)

        # find the user/assistant pairs
        messages = []
        found = True
        while found:

            # find the user text
            found = False
            for role_name,start_bytes in (("system", b'<|im_start|>system\n'), ("user", b'<|im_start|>user\n'), ("assistant", b'<|im_start|>assistant\n')):
                if prompt[pos:].startswith(start_bytes):
                    pos += len(start_bytes)
                    end_pos = prompt[pos:].find(role_end)
                    if end_pos < 0:
                        assert role_name == "assistant", "Bad chat format! Last role before gen needs to be assistant!"
                        break
                    btext = prompt[pos:pos+end_pos]
                    pos += end_pos + len(role_end)
                    messages.append({"role": role_name, "content": btext.decode("utf8")})
                    found = True
                    break
        
        self._shared_state["data"] = prompt[:pos]

        try:
            generator = self.client.chat.completions.create(
                model=self.model_name,
                messages=messages,
                max_tokens=self.max_streaming_tokens, 
                n=1, 
                top_p=1, 
                temperature=0, 
                stream=True
            )
        except Exception as e: # TODO: add retry logic
            raise e

        for part in generator:
            chunk = part.choices[0].delta.content or ""
            yield chunk.encode("utf8")


from .._transformers import Transformers

class LLaMA(Transformers):
    """ A HuggingFace transformers version of the LLaMA language model with Guidance support.
    """

    llm_name: str = "llama"

    def _model_and_tokenizer(self, model, tokenizer, **kwargs):

        # load the LLaMA specific tokenizer and model
        import transformers
        if isinstance(model, str):
            if tokenizer is None:
                tokenizer = transformers.LlamaTokenizer.from_pretrained(model, **kwargs)
            model = transformers.LlamaForCausalLM.from_pretrained(model, **kwargs)
            
        return super()._model_and_tokenizer(model, tokenizer, **kwargs)

class LLaMAChat(LLaMA):

    default_system_prompt = """You are a helpful assistant."""

    # https://github.com/facebookresearch/llama/blob/main/llama/generation.py#L45
    # This is not exatly right, I think they do a single [INST] for the system prompt AND first instruction
    @staticmethod
    def role_start(role):
       return  {
        'system': '[INST] <<SYS>>\n',
        'user': '[INST] ',
        'assistant': '',
        }[role]
    
    @staticmethod
    def role_end(role):
       return  {
        'system': '\n<</SYS>>\n\n[/INST]',
        'user': '[/INST]',
        'assistant': '',
        }[role]


import guidance
from guidance import gen, models, commit_point, Tool, select, capture, string, substring
from ..utils import get_model
import re

def test_substring_equal_unconstrained():
    llama2 = get_model("llama_cpp:")
    lm = llama2 + 'ae galera ' + gen(max_tokens=10, name='test')
    lm2 = llama2 + 'ae galera ' + substring(lm['test'])
    assert str(lm) == str(lm2)


import guidance
from guidance import gen, capture
from ..utils import get_model

def test_gpt2():
    gpt2 = get_model("transformers:gpt2")
    lm = gpt2 + "this is a test" + gen("test", max_tokens=10)
    assert len(str(lm)) > len("this is a test")

def test_recursion_error():
    '''This is similar an infinite recursion error resulting from a terminal node at the root of a grammar.'''
    gpt2 = get_model("transformers:gpt2")

    # define a guidance program that adapts a proverb
    lm = gpt2 + f"""Tweak this proverb to apply to model instructions instead.
    {gen('verse', max_tokens=2)}
    """
    assert len(str(lm)) > len("Tweak this proverb to apply to model instructions instead.\n\n")


from guidance import models, select
from .utils import get_model

def test_select_reset_pos():
    model = models.LocalMock()
    model += 'This is' + select(options=['bad', 'quite bad'])
    assert str(model) in ["This isbad", "This isquite bad"]

def test_select_simple():
    lm = get_model("transformers:gpt2")
    options = ['baad I think', 'bad I think', 'bad']
    lm = lm + 'Scott is quite ' + select(name='bad', options=options)
    assert lm['bad'] in options

def test_select_longer():
    '''This tests to ensure that the grammar is extended greedily.'''
    lm = models.LocalMock(b"<s>Scott is a very nice man.")
    lm += "Scott is a very " + select(name='text', options=['nice', 'nice man.'])
    assert lm["text"] == 'nice man.'



from guidance import one_or_more, zero_or_more, char_set, string, select
from guidance._grammar import Byte, ByteRange
from guidance._parser import EarleyCommitParser

def test_one_or_more():
    g = one_or_more("a")
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b'a')])
    parser.consume_byte(b'a')
    assert parser.valid_next_bytes() == set([Byte(b'a')])

def test_zero_or_more_and_one_or_more():
    g = zero_or_more("a") + one_or_more("b")
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b'a'), Byte(b'b')])
    parser.consume_byte(b'a')
    assert parser.valid_next_bytes() == set([Byte(b'a'), Byte(b'b')])
    parser.consume_byte(b'b')
    assert parser.valid_next_bytes() == set([Byte(b'b')])

    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b'a'), Byte(b'b')])
    parser.consume_byte(b'b')
    assert parser.valid_next_bytes() == set([Byte(b'b')])
    parser.consume_byte(b'b')
    assert parser.valid_next_bytes() == set([Byte(b'b')])

def test_zero_or_more_and_one_or_more_mixed():
    g = zero_or_more("a") + "test" + one_or_more("b")
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b'a'), Byte(b't')])
    parser.consume_byte(b't')
    parser.consume_byte(b'e')
    parser.consume_byte(b's')
    assert parser.valid_next_bytes() == set([Byte(b't')])
    parser.consume_byte(b't')
    assert parser.valid_next_bytes() == set([Byte(b'b')])

def test_select():
    g = select(["bob", "bill", "sue"])
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b'b'), Byte(b's')])
    parser.consume_byte(b's')
    assert parser.valid_next_bytes() == set([Byte(b'u')])
    parser.consume_byte(b'u')
    assert parser.valid_next_bytes() == set([Byte(b'e')])

def test_select_nested():
    g = select(["bob", "bill", select(["mark", "mary"])])
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b'b'), Byte(b'm')])
    parser.consume_byte(b'm')
    assert parser.valid_next_bytes() == set([Byte(b'a')])
    parser.consume_byte(b'a')
    assert parser.valid_next_bytes() == set([Byte(b'r')])
    parser.consume_byte(b'r')
    assert parser.valid_next_bytes() == set([Byte(b'k'), Byte(b'y')])

def test_select_joined():
    g = select(["bob", "bill"]) + select(["mark", "mary"])
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b'b')])
    parser.consume_byte(b'b')
    assert parser.valid_next_bytes() == set([Byte(b'o'), Byte(b'i')])
    parser.consume_byte(b'i')
    assert parser.valid_next_bytes() == set([Byte(b'l')])
    parser.consume_byte(b'l')
    assert parser.valid_next_bytes() == set([Byte(b'l')])
    parser.consume_byte(b'l')
    assert parser.valid_next_bytes() == set([Byte(b'm')])
    parser.consume_byte(b'm')
    assert parser.valid_next_bytes() == set([Byte(b'a')])
    parser.consume_byte(b'a')
    assert parser.valid_next_bytes() == set([Byte(b'r')])
    parser.consume_byte(b'r')
    assert parser.valid_next_bytes() == set([Byte(b'k'), Byte(b'y')])

def test_char_set():
    g = char_set("b-f")
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([ByteRange(b'bf')])
    parser.consume_byte(b'b')

def test_byte_mask_char_set():
    g = char_set("b-f")
    parser = EarleyCommitParser(g)
    m = parser.next_byte_mask()
    for i in range(256):
        if ord(b'b') <= i <= ord(b'f'):
            assert m[i]
        else:
            assert not m[i]

def test_byte_mask_char_set2():
    g = char_set("bf")
    parser = EarleyCommitParser(g)
    m = parser.next_byte_mask()
    for i in range(256):
        if i == ord(b'b') or i == ord(b'f'):
            assert m[i]
        else:
            assert not m[i]

def test_char_set_one_or_more():
    g = one_or_more(char_set("b-f"))
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([ByteRange(b'bf')])
    parser.consume_byte(b'b')
    assert parser.valid_next_bytes() == set([ByteRange(b'bf')])
    parser.consume_byte(b'b')
    assert parser.valid_next_bytes() == set([ByteRange(b'bf')])
    parser.consume_byte(b'f')
    assert parser.valid_next_bytes() == set([ByteRange(b'bf')])

def test_string_utf8():
    b = bytes("¶", encoding="utf8")
    g = string("¶")
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b[:1])])
    parser.consume_byte(b[:1])
    assert parser.valid_next_bytes() == set([Byte(b[1:])])
    parser.consume_byte(b[1:])


from ._model import Model, Chat
from .vertexai._vertexai import VertexAI, VertexAIChat
from ._openai import OpenAI, OpenAIChat, OpenAIInstruct, OpenAICompletion
from .transformers._transformers import Transformers, TransformersChat
from ._llama_cpp import LlamaCpp, LlamaCppChat
from ._local_mock import LocalMock, LocalMockChat
from . import transformers


from guidance import models, any_char_but

def test_single_char():
    model = models.LocalMock("<s>abc")
    assert str(model + '<s>' + any_char_but('a')) != "<s>a"
    assert str(model + '<s>' + any_char_but('!')) == "<s>a"

def test_multi_char():
    model = models.LocalMock(["<s>abc", "<s>bbc"])
    assert str(model + '<s>' + any_char_but('ab')) not in ("<s>a", "<s>b")
    assert str(model + '<s>' + any_char_but('a!')) == "<s>b"
    assert str(model + '<s>' + any_char_but('5b')) == "<s>a"


import types
import regex as lregex
import uuid

import guidance
import ast

# from guidance import select, any_char, zero_or_more, commit_point, hide
from ._silent import silent
from .._grammar import select
from ._zero_or_more import zero_or_more
from .._grammar import commit_point
from ._any_char import any_char
from .._grammar import capture
from ._regex import regex as regex_grammar
from .._grammar import token_limit
from .._grammar import with_temperature
from .._grammar import model_variable
from ._tool import Tool

# TODO: make this stateless!
@guidance(stateless=lambda *args, **kwargs: kwargs.get("tools", None) is None) # TODO: uncomment this once we get temperature stateless
def gen(lm, name=None, *, max_tokens=1000, list_append=False, regex=None,
        tools=None, hide_tool_call=False, stop=None, stop_regex=None, suffix="", n=1, temperature=0.0, top_p=1.0,
        logprobs=None, stream_tokens=None, save_stop_text=False, **llm_kwargs):
    """
    TODO: document this
    tools is a list of guidance.Tool or python functions (which will be converted to guidance.Tool)
    """

    # set stream if we are interactive
    # if stream_tokens is None and not lm.is_silent() and n == 1:
    #     stream_tokens = True

    # use the suffix as the stop string if not otherwise specified
    # TODO: still need to make suffix work with grammars
    # eos_token = lm.eos_token.decode('utf8')
    if stop is None and stop_regex is None and suffix != "":
        stop = suffix
    # if stop is None and stop_regex is None and getattr(lm, "suffix", False):
    #     if lm.suffix.startswith("\n"):
    #         stop = "\n"
    #     elif lm.suffix.startswith('"') and str(lm).endswith('"'):
    #         stop = '"'
    #     elif lm.suffix.startswith("'") and str(lm).endswith("'"):
    #         stop = "'"

    # fall back to stopping at the EOS token
    if stop is None:
        stop = []
    if isinstance(stop, str):
        stop = [stop]
    if regex is None:
        stop.append(model_variable('default_end_patterns'))

    if stop_regex is None:
        stop_regex = []
    if isinstance(stop_regex, str):
        stop_regex = [stop_regex]
    stop_regex = [regex_grammar(x) for x in stop_regex]

    # This needs to be here for streaming
    # if name is not None and not list_append:
    #     lm[name] = ""

    # TODO: This can be uncommented once we support regex -> grammar compilation
    
    # compile stop_regex into a capture group
    # if "(?P<stop>" in pattern:
    #     assert stop_regex == None, "You can't specify both a stop string/regex and a custom stop pattern (?P<stop>...)!"
    #     stop_regex = ""
    # else:
    #     if stop_regex is None:
    #         stop_regex = []
    #     stop_regex.append(regex.escape(lm.eos_token))
    #     stop_regex = "(?P<stop>" + "|".join(stop_regex) + ")"

    # pattern += stop_regex
    # extracted_stop_pattern = regex.compile(pattern[pattern.index("(?P<stop>")+9:-1] + "$", flags=regex.DOTALL)
    # extracted_stop_pattern = regex.compile(pattern[:pattern.index("(?P<stop>")] + "$", flags=regex.DOTALL)
    
    # define the generation pattern
    if regex is not None:
        pattern = regex_grammar(regex)
    else:
        pattern = zero_or_more(any_char())

    # define any capture group
    if name is not None:
        pattern = capture(pattern, name="__LIST_APPEND:" + name if list_append else name)
    
    # limit the number of tokens
    pattern = token_limit(pattern, max_tokens)
    
    # define the stop pattern
    if stop + stop_regex:
        stop_pattern = select(stop + stop_regex)
        if save_stop_text is True:
            save_stop_text = str(name) + "_stop_text"
        if isinstance(save_stop_text, str):
            stop_pattern = capture(stop_pattern, name=save_stop_text)
        stop_pattern = commit_point(stop_pattern, hidden=True)
    else:
        stop_pattern = ''

    # single generation
    start_pos = len(str(lm))
    # TODO: if a tool is a python function rather than a guidance.Tool, make it into a guidance.Tool
    if tools is not None:
        # TODO: This should be while I have tokens left
        tools = [Tool(callable=x) if not isinstance(x, Tool) else x for x in tools]
        gen_grammar = pattern + select([stop_pattern] + [capture(commit_point(x.call_grammar, hidden=hide_tool_call), name=f'tool{i}') for i, x in enumerate(tools)])
        while lm._token_count <= max_tokens:
        # for i in range(5):
            lm = lm.run_stateless(gen_grammar, temperature=temperature)
            tool_called = False
            for i in range(len(tools)):
                tool_i = f'tool{i}'
                if tool_i in lm:
                    tool_called = True
                    lm += tools[i].tool_call()
                    lm.remove(tool_i)
            if not tool_called:
                lm += suffix
                break
    elif n == 1:
        lm += with_temperature(pattern + stop_pattern + suffix, temperature)

    return lm


def click_loop_start(id, total_count, echo, color):
    click_script = '''
function cycle_IDVAL(button_el) {
var i = 0;
while (i < 50) {
var el = document.getElementById("IDVAL_" + i);
if (el.style.display == "inline") {
    el.style.display = "none";
    var next_el = document.getElementById("IDVAL_" + (i+1));
    if (!next_el) {
        next_el = document.getElementById("IDVAL_0");
    }
    if (next_el) {
        next_el.style.display = "inline";
    }
    break;
}
i += 1;
}
button_el.innerHTML = (((i+1) % TOTALCOUNT) + 1)  + "/" + TOTALCOUNT;
}
cycle_IDVAL(this);'''.replace("IDVAL", id).replace("TOTALCOUNT", str(total_count)).replace("\n", "")
    out = f'''<div style='background: rgba(255, 255, 255, 0.0); border-radius: 4px 0px 0px 4px; border: 1px solid {color}; border-right: 0px; padding-left: 3px; padding-right: 3px; user-select: none; color: {color}; display: inline; font-weight: normal; cursor: pointer' onClick='{click_script}'>1/{total_count}</div>'''
    out += f"<div style='display: inline;' id='{id}_0'>"
    return "<||_html:" + out + "_||>"

def click_loop_mid(id, index, echo):
    alpha = 1.0 if not echo else 0.5
    out = f"</div><div style='display: none; opacity: {alpha}' id='{id}_{index}'>"
    return "<||_html:" + out + "_||>"

@guidance
def gen_line(lm, *args, **kwargs):
    return lm.gen(*args, suffix='\n', **kwargs)

@guidance
def gen_quote(lm, name=None, quote='"', *args, **kwargs):
    return lm(quote).gen(*args,name=name, suffix=quote, **kwargs)

@guidance
def will_gen(lm, stop=None, stop_regex=None, ignore_spaces=False, max_tokens=30):
    # this is obviously not the right implementation, just here so we can explore
    if stop and not isinstance(stop, list):
        stop = [stop]
    if stop_regex and not isinstance(stop_regex, list):
        stop_regex = [stop_regex]
    assert (stop is not None) or (stop_regex is not None)
    if not stop:
        stop = []
    if not stop_regex:
        stop_regex = []
    regexes = [lregex.escape(x) for x in stop + stop_regex]
    optional_space = '\\s*' if ignore_spaces else ''
    pattern = lregex.compile(f'{optional_space}({"|".join(regexes)})')
    lm2 = lm
    with silent():
        for _ in range(max_tokens):
            lm2 += gen('temp_variable', list_append=True, max_tokens=1)
            if not lm2['temp_variable'] or not pattern.match(''.join(lm2['temp_variable']), partial=True):
                return False
            if pattern.match(''.join(lm2['temp_variable']), partial=False):
                return True
    return False

@guidance
def call_tool(lm, tool):
    return lm + tool.call_grammar + tool.tool_call()